
from typing import (
    Any,
    Union, Optional,
    Tuple, List, Set, Sequence, Dict,
    Awaitable, AsyncGenerator, Coroutine,
)
import collections
import datetime
import json
import asyncio
from http.client import responses as RESPONSE_CODES

import httpx

from .models import Message, System, Member, Switch, Timestamp
from .errors import *
from .utils import *

SERVER = "https://api.pluralkit.me/v1"
RATE_LIMIT_THROTTLE = 0.1 # seconds

class Client:
    """Represents a client that interacts with the PluralKit API.

    Args:
        token: The PluralKit authorization token, received by the ``pk;token`` command.

    Keyword args:
        async_mode: Whether the client runs asynchronously (``True``, default) or not (``False``).
        user_agent: The User-Agent header to use with the API.

    Attributes:
        token (Optional[str]): The client's PluralKit authorization token.
        async_mode (bool): Whether the client runs asynchronously (``True``) or not (``False``).
        user_agent (Optional[str]): The User-Agent header used with the API.
        id (Optional[str]): The five-letter lowercase ID of one's system if an authorization token
            is provided.
    """
    def __init__(self, token: Optional[str]=None, *,
        async_mode: bool=True,
        user_agent: Optional[str]=None
    ):
        self._calls_queue = collections.deque()
        self.async_mode = async_mode
        self.token = token
        self.headers = {}
        self.id = None
        if token:
            self.headers["Authorization"] = token
            if async_mode:
                loop = asyncio.get_event_loop()
                system = loop.run_until_complete(self._get_system())
            else:
                system = self.get_system() # type: ignore
            self.id = system.id
        self.user_agent = user_agent
        if user_agent:
            self.headers["User-Agent"] = user_agent
        self.content_headers = self.headers.copy()
        self.content_headers["Content-Type"] = "application/json"

    def _num_calls_in_last_half_second(self):
        half_second = datetime.timedelta(seconds=0.5)
        while len(self._calls_queue) \
                and datetime.datetime.now() - self._calls_queue[-1] > half_second:
            self._calls_queue.pop()

        return len(self._calls_queue)

    async def _respect_rate_limit(self):
        while self._num_calls_in_last_half_second() >= 1:
            await asyncio.sleep(RATE_LIMIT_THROTTLE)

        self._calls_queue.append(datetime.datetime.now())

    def get_system(self, system: Union[System,str,int,None]=None) \
    -> Union[System, Coroutine[Any,Any,System]]:
        """Return a system by its system ID or Discord user ID.

        Args:
            system: The system ID (as `str`), Discord user ID (as `int`), or `System` object of the
                system. If ``None``, returns the system of the client.

        Returns:
            System: The retrieved system.
        """
        awaitable = self._get_system(system)
        if self.async_mode:
            return awaitable
        else:
            loop = asyncio.get_event_loop()
            result = loop.run_until_complete(awaitable)
            return result

    async def _get_system(self, system: Union[System,str,int,None]=None) -> System:
        if system is None:
            if not self.token: raise AuthorizationError() # please pass in your token to the client
            # get own system
            url = f"{SERVER}/s"
        elif isinstance(system, System):
            # System object
            url = f"{SERVER}/s/{system.id}"
        elif isinstance(system, str):
            # system ID
            url = f"{SERVER}/s/{system}"
        elif isinstance(system, int):
            # Discord user ID
            url = f"{SERVER}/a/{system}"

        await self._respect_rate_limit()

        async with httpx.AsyncClient(headers=self.headers) as session:
            response = await session.get(url)
            if response.status_code == 401:
                raise AuthorizationError()
            elif response.status_code == 404:
                if isinstance(system, str):
                    raise SystemNotFound(system)
                elif isinstance(system, int):
                    raise DiscordUserNotFound(system)
                
                if response.status_code != 200: # catch-all
                    raise HTTPError(response.status_code)

            resp = response.json()
            new_system = System.from_json(resp)

            return new_system
    
    def edit_system(self, system: Optional[System]=None, **kwargs) \
    -> Union[System, Coroutine[Any,Any,System]]:
        """Edits one's own system

        Note:
            The system's `authorization token`_ must be set in order to use this method.

        If a `System` object is provided, this method will use the attributes of that
        object to edit one's system. If *both* a System object and some keyword arguments are
        passed, the keyword arguments will take priority.

        Args:
            system: The optional object to transfer system attributes from.
            **kwargs: Any number of keyworded patchable values from `PluralKit's system model`_.

        Note:
            Any fields not passed will retain their prior value. Any fields passed with ``None``
            will clear or reset to the default.

        Keyword Args:
            name (Optional[str]): The new system name.
            description (Optional[str]): The new system description.
            tag (Optional[str]): The new system tag.
            avatar_url (Optional[str]): The new system avatar.
            tz (Union[Timezone,str,None]): The timezone as a tzdb identifier. Passing `None` will
                reset this field to "UTC".
            description_privacy (Union[Privacy,str,None]): The new system description privacy.
            member_list_privacy (Union[Privacy,str,None]): The new system member list privacy.
            front_privacy (Union[Privacy,str,None]): The new system front privacy.
            metadata_privacy (Union[Privacy,str,None]): The new system metadata privacy.

        Returns:
            System: The updated system.

        .. _`PluralKit's system model`: https://pluralkit.me/api/#system-model
        .. _`authorization token`: https://pluralkit.me/api/#authentication
        """
        awaitable = self._edit_system(system, **kwargs)
        if self.async_mode:
            return awaitable
        else:
            loop = asyncio.get_event_loop()
            result = loop.run_until_complete(awaitable)
            return result

    async def _edit_system(self, system: Optional[System]=None, **kwargs) -> System:
        if self.token is None:
            raise AuthorizationError() # catches a 401 before it happens

        if system is not None:
            kwargs_ = system.json()
            kwargs_.update(kwargs)
            kwargs = kwargs_

        if "id" in kwargs:
            del kwargs["id"]
        if "created" in kwargs:
            del kwargs["created"]

        for key, value in kwargs.items():
            await system_value(key=key, value=value)
        
        payload = json.dumps(kwargs, ensure_ascii=False)

        await self._respect_rate_limit()

        async with httpx.AsyncClient(headers=self.content_headers) as session:
            response = await session.patch(f"{SERVER}/s", data=payload)
            if response.status_code != 200: # catch-all
                raise HTTPError(response.status_code)

            resp = response.json()
            system = System.from_json(resp)
            return system

    def get_fronters(self, system=None) \
    -> Union[Tuple[Timestamp, List[Member]], Coroutine[Any,Any,Tuple[Timestamp, List[Member]]]]:
        """Fetches the current fronters of a system.
        
        Args:
            system (Optional[Union[str,System,int]]): The system to fetch fronters from. Can be a
                System object, the five-letter lowercase system ID as a string, or the Discord user
                ID corresponding to a system. If ``None``, fetches the fronters of the system
                associated with the client.

        Note:
            Passing a Discord user ID takes an extra request. For this reason, it's perhaps better
            practice to cache the system ID associated with Discord accounts using
            `Client.get_system` first.
        
        Returns:
            Tuple[Timestamp, List[Member]]: The Timestamp object and the list of current fronters.
        """
        awaitable = self._get_fronters(system)
        if self.async_mode:
            return awaitable
        else:
            loop = asyncio.get_event_loop()
            result = loop.run_until_complete(awaitable)
            return result

    async def _get_fronters(self, system=None) -> Tuple[Timestamp, List[Member]]:
        async with httpx.AsyncClient(headers=self.headers) as session:
            if system is None:
                # get own system
                url = f"{SERVER}/s/{self.id}/fronters"
            elif isinstance(system, System):
                # System object
                url = f"{SERVER}/s/{system.id}/fronters"
            elif isinstance(system, str):
                # system ID
                url = f"{SERVER}/s/{system}/fronters"
            elif isinstance(system, int):
                # Discord user ID

                await self._respect_rate_limit()
                response = await session.get(f"{SERVER}/a/{system}")

                if response.status_code == 401:
                    raise AuthorizationError()

                if response.status_code == 404:
                    raise DiscordUserNotFound(system)
                    
                if response.status_code != 200: # catch-all
                    raise HTTPError(response.status_code)

                system_ = System.from_json(response.json())
                print(system_.json())
                url = f"{SERVER}/s/{system_.id}/fronters"
                print(url)

            await self._respect_rate_limit()
            response = await session.get(url)

            if response.status_code == 401:
                raise AuthorizationError()

            if response.status_code == 403:
                raise AccessForbidden()

            if response.status_code == 404:
                if isinstance(system, System):
                    raise SystemNotFound(system.id)
                raise SystemNotFound(system)

            if response.status_code != 200: # catch-all
                raise HTTPError(response.status_code)

            resp = response.json()
            member_list = []
            for fronter in resp["members"]:
                member_list.append(Member.from_json(fronter))
            timestamp = Timestamp.from_json(resp["timestamp"])
            return (timestamp, member_list)

    def get_members(self, system: Union[System,str,int,None]=None
    ) -> Union[List[Member], AsyncGenerator[Member,None]]:
        """Retrieve list of a system's members.

        Args:
            system (Optional[Union[str,System,int]]): The system to fetch members from. Can be a
                System object, the five-letter lowercase system ID as a string, or the Discord user
                ID corresponding to a system. If ``None``, fetches the members of the system
                associated with the client.

        Note:
            Passing a Discord user ID takes an extra request. For this reason, it's perhaps better
            practice to cache the system ID associated with Discord accounts using
            `Client.get_system` first.

        Yields:
            Member: The next system member.
        """
        awaitable = self._get_members(system)
        if self.async_mode:
            return awaitable
        else:
            loop = asyncio.get_event_loop()
            result = loop.run_until_complete(flatten(awaitable))
            return result

    async def _get_members(self, system: Union[System,str,int,None]=None) \
    -> AsyncGenerator[Member,None]:

        async with httpx.AsyncClient(headers=self.headers) as session:
            if system is None:
                # get own system
                url = f"{SERVER}/s/{self.id}/members"
            elif isinstance(system, System):
                # System object
                url = f"{SERVER}/s/{system.id}/members"
            elif isinstance(system, str):
                # system ID
                url = f"{SERVER}/s/{system}/members"
            elif isinstance(system, int):
                # Discord user ID

                await self._respect_rate_limit()
                response = await session.get(f"{SERVER}/a/{system}")

                if response.status_code == 401:
                    raise AuthorizationError()

                if response.status_code == 404:
                    raise DiscordUserNotFound(system)
                    
                if response.status_code != 200: # catch-all
                    raise HTTPError(response.status_code)

                system_ = System.from_json(response.json())
                url = f"{SERVER}/s/{system_.id}/members"

            await self._respect_rate_limit()
            response = await session.get(url)

            if response.status_code == 401:
                raise AuthorizationError()

            if response.status_code == 403:
                raise AccessForbidden()

            if response.status_code == 404:
                if isinstance(system, System):
                    raise SystemNotFound(system.id)
                raise SystemNotFound(system)
                
            if response.status_code != 200: # catch-all
                raise HTTPError(response.status_code)

            resp = response.json()

            for item in resp:
                member = Member.from_json(item)

                yield member
    
    def get_member(self, member_id: str) -> Union[Member, Coroutine[Any,Any,Member]]:
        """Gets a system member.

        Args:
            member_id: The ID of the member to be fetched.

        Returns:
            Member: The member with the given ID.
        """
        awaitable = self._get_member(member_id)
        if self.async_mode:
            return awaitable
        else:
            loop = asyncio.get_event_loop()
            result = loop.run_until_complete(awaitable)
            return result

    async def _get_member(self, member_id: str) -> Member:
        async with httpx.AsyncClient(headers=self.headers) as session:
            await self._respect_rate_limit()
            response = await session.get(f"{SERVER}/m/{member_id}")

            if response.status_code == 404:
                raise MemberNotFound(member_id)

            if response.status_code != 200:
                raise HTTPError(response.status_code)
            
            resp = response.json()
            return Member.from_json(resp)

    def new_member(self, name: str, **kwargs) -> Union[Member, Coroutine[Any,Any,Member]]:
        """Creates a new member of one's system.

        Note:
            The system's `authorization token`_ must be set in order to use this method.

        Args:
            name: Name of the new member.
            **kwargs: Any number of keyworded patchable values from `PluralKit's member model`_.
        
        Keyword Args:
            display_name (Optional[str]): New display name of the member. Default is None.
            description (Optional[str]): New description of the member. Default is None.
            pronouns (Optional[str]): New pronouns of the member. Default is None.
            color (Union[Color,str,None]): New color of the member. If a string, must be formatted
                as a 6-character hex string (e.g. ``"ff7000"``), sans the # symbol. Default is
                ``None``.
            avatar_url (Optional[str]): New avatar URL for the member. Default is None.
            birthday (Union[Timestamp,str]): New birthdate of the member. If a string, must be
                formatted as ``YYYY-MM-DD``, in which case, year of ``0001`` or ``0004`` represents
                a hidden year. Default is None.
            proxy_tags (Union[ProxyTags,Sequence[ProxyTag],Sequence[Dict[str,str]]]): New proxy
                tags of the member. May be a `ProxyTags` object, a sequence of `ProxyTag` objects,
                or a sequence of Python dictionaries with the keys "prefix" and "suffix". Default
                is an empty set of proxy tags.
            keep_proxy (bool): New truth value for whether to display the member's proxy tags in
                the proxied message. Default is ``False``.
            visibility (Union[Privacy,str,None]): New visibility privacy for the member. Must be
                either `Privacy.PUBLIC` or `Privacy.PRIVATE`. Default is public.
            name_privacy (Union[Privacy,str,None]): New name privacy for the member. Must be either
                `Privacy.PUBLIC` or `Privacy.PRIVATE`. Default is public.
            description_privacy (Union[Privacy,str,None]): New description privacy for the member.
                Must be either `Privacy.PUBLIC` or `Privacy.PRIVATE`. Default is public.
            avatar_privacy (Union[Privacy,str,None]): New avatar privacy for the member. Must be
                either `Privacy.PUBLIC` or `Privacy.PRIVATE`. Default is public.
            pronoun_privacy (Union[Privacy,str]): New pronouns privacy for the member. Must be
                either `Privacy.PUBLIC` or `Privacy.PRIVATE`. Default is public.
            metadata_privacy (Union[Privacy,str]): New metadata (eg. creation timestamp, message
                count, etc.) privacy for the member. Must be either `Privacy.PUBLIC` or
                `Privacy.PRIVATE`. Default is public.

        Returns:
            Member: The new member.

        .. _`PluralKit's member model`: https://pluralkit.me/api/#member-model
        .. _`authorization token`: https://pluralkit.me/api/#authentication
        """
        awaitable = self._new_member(name, **kwargs)
        if self.async_mode:
            return awaitable
        else:
            loop = asyncio.get_event_loop()
            result = loop.run_until_complete(awaitable)
            return result

    async def _new_member(self, name: str, **kwargs) -> Member:
        if self.token is None:
            raise AuthorizationError()

        kwargs["name"] = name

        for key, value in kwargs.items():
            kwargs = await member_value(kwargs=kwargs, key=key, value=value)
        
        payload = json.dumps(kwargs, ensure_ascii=False)

        async with httpx.AsyncClient(headers=self.content_headers) as session:
            
            await self._respect_rate_limit()
            response = await session.post(f"{SERVER}/m/", data=payload)

            if response.status_code == 401:
                raise AuthorizationError

            if response.status_code != 200:
                raise HTTPError(response.status_code)

            resp = response.json()
            return Member.from_json(resp)

    def edit_member(self, 
                    member_id: Union[str, Member], 
                    member: Optional[Member]=None, 
                    **kwargs) -> Union[Member, Coroutine[Any,Any,Member]]:
        """Edits a member of one's system.

        Note:
            The system's `authorization token`_ must be set in order to use this method.

        If a `Member` object is provided, this method will use the attributes of that
        object to edit the member. If *both* a Member object and some keyword arguments are
        passed, the keyword arguments will take priority.

        Args:
            member_id: The member to be edited. Can be a Member object or the member ID as a
                string.
            member: Optional Member object representing the member with updated attributes.
            **kwargs: Any number of keyworded patchable values from `PluralKit's member model`_.
        
        Note:
            Any fields not passed will retain their prior value. Any fields passed with ``None``
            will clear or reset to the default.

        Keyword Args:
            name (str): New name of the member.
            display_name (Optional[str]): New display name of the member. If ``None`` is passed,
                this field is cleared.
            description (Optional[str]): New description of the member. If ``None`` is passed, this
                field is cleared.
            pronouns (Optional[str]): New pronouns of the member. If ``None`` is passed, this field
                is cleared.
            color (Union[Color,str,None]): New color of the member. If a string, must be formatted
                as a 6-character hex string (e.g. ``"ff7000"``), sans the # symbol. If ``None`` is
                passed, this field is cleared.
            avatar_url (Optional[str]): New avatar URL for the member. If ``None`` is passed, this
                field is cleared.
            birthday (Union[Birthday,str]): New birthdate of the member. If a string, must be
                formatted as ``YYYY-MM-DD``, in which case, a year of ``0001`` or ``0004``
                represents a hidden year. If ``None`` is passed, this field is cleared.
            proxy_tags (Union[ProxyTags,Sequence[ProxyTag],Sequence[Dict[str,str]]]): New proxy
                tags of the member. May be a `ProxyTags` object, a sequence of `ProxyTag` objects,
                or a sequence of Python dictionaries with the keys "prefix" and "suffix".
            keep_proxy (bool): New truth value for whether to display the member's proxy tags in
                the proxied message.
            visibility (Union[Privacy,str,None]): New visibility privacy for the member. Must be
                either `Privacy.PUBLIC` or `Privacy.PRIVATE`. If ``None`` is passed, this field is
                reset to public.
            name_privacy (Union[Privacy,str,None]): New name privacy for the member. Must be either
                `Privacy.PUBLIC` or `Privacy.PRIVATE`. If ``None`` is passed, this field is reset
                to public.
            description_privacy (Union[Privacy,str,None]): New description privacy for the member.
                Must be either `Privacy.PUBLIC` or `Privacy.PRIVATE`. If ``None`` is passed, this
                field is reset to public.
            avatar_privacy (Union[Privacy,str,None]): New avatar privacy for the member. Must be
                either `Privacy.PUBLIC` or `Privacy.PRIVATE`. If ``None`` is passed, this field is
                reset to public.
            pronoun_privacy (Union[Privacy,str]): New pronouns privacy for the member. Must be
                either `Privacy.PUBLIC` or `Privacy.PRIVATE`. If ``None`` is passed, this field is
                reset to public.
            metadata_privacy (Union[Privacy,str]): New metadata (eg. creation timestamp, message
                count, etc.) privacy for the member. Must be either `Privacy.PUBLIC` or
                `Privacy.PRIVATE`. If ``None`` is passed, this field is reset to public.

        Returns:
            Member: The updated member.

        .. _`PluralKit's member model`: https://pluralkit.me/api/#member-model
        .. _`authorization token`: https://pluralkit.me/api/#authentication
        """
        
        awaitable = self._edit_member(member_id, member, **kwargs)
        if self.async_mode:
            return awaitable
        else:
            loop = asyncio.get_event_loop()
            result = loop.run_until_complete(awaitable)
            return result

    async def _edit_member(self, member_id: str, member: Optional[Member]=None, **kwargs) -> Member:
        if self.token is None:
            raise AuthorizationError()

        if isinstance(member, Member):
            kwargs_ = member.json()
            kwargs_.update(kwargs)
            kwargs = kwargs_

        if "id" in kwargs:
            del kwargs["id"]
        if "created" in kwargs:
            del kwargs["created"]

        for key, value in kwargs.items():
            kwargs = await member_value(kwargs=kwargs, key=key, value=value)
        
        payload = json.dumps(kwargs, ensure_ascii=False)
        
        async with httpx.AsyncClient(headers=self.content_headers) as session:
            
            await self._respect_rate_limit()
            response = await session.patch(f"{SERVER}/m/{member_id}", data=payload)
            
            if response.status_code == 401:
                raise AuthorizationError

            if response.status_code != 200:
                raise HTTPError(response.status_code)

            resp = response.json()
            return Member.from_json(resp)

    def delete_member(self, member_id: Union[str,Member]) \
    -> Union[None, Coroutine[Any,Any,None]]:
        """Deletes a member of one's system
        
        Note:
            The system's `authorization token`_ must be set in order to use this method.
        
        Args:
            member_id: The member to be deleted. Can be a Member object or the member ID as a
                string.

        Returns:
            `NoneType`: None

        .. _`authorization token`: https://pluralkit.me/api/#authentication
        """
        awaitable = self._delete_member(member_id)
        if self.async_mode:
            return awaitable
        else:
            loop = asyncio.get_event_loop()
            result = loop.run_until_complete(awaitable)
            return result

    async def _delete_member(self, member_id: Union[str,Member]) -> None:
        url = f"{SERVER}/m/{member_id}"

        async with httpx.AsyncClient(headers=self.headers) as session:
    
            await self._respect_rate_limit()
            response = await session.delete(url)

            if response.status_code == 401:
                raise AuthorizationError()

            if response.status_code == 403:
                raise AccessForbidden()

            if response.status_code != 200: # catch-all
                raise HTTPError(response.status_code)

        return None

    def get_switches(self, system: Optional[Union[System,str]]=None) \
    -> Union[List[Switch], AsyncGenerator[Switch,None]]:
        """Fetches the switch history of a system.
        
        Args:
            system: The system to fetch switch history from. Can be a System object, the
                five-letter lowercase system ID as a string, or the Discord user ID corresponding
                to a system. If ``None``, fetches the switch history of the system associated with
                the client.

        Note:
            Passing a Discord user ID takes an extra request. For this reason, it's perhaps better
            practice to cache the system ID associated with Discord accounts using
            `Client.get_system` first.
            
        Yields:
            Switch: The next switch.
        """
        awaitable = self._get_switches(system)
        if self.async_mode:
            return awaitable
        else:
            loop = asyncio.get_event_loop()
            result = loop.run_until_complete(flatten(awaitable))
            return result
        
    async def _get_switches(self, system=None) -> AsyncGenerator[Switch,None]:
        async with httpx.AsyncClient(headers=self.headers) as session:
            if system is None:
                # get own system
                url = f"{SERVER}/s/{self.id}/switches"
            elif isinstance(system, System):
                # System object
                url = f"{SERVER}/s/{system.id}/switches"
            elif isinstance(system, str):
                # system ID
                url = f"{SERVER}/s/{system}/switches"
            elif isinstance(system, int):
                # Discord user ID

                await self._respect_rate_limit()
                response = await session.get(f"{SERVER}/a/{system}")

                if response.status_code == 401:
                    raise AuthorizationError()

                if response.status_code == 404:
                    raise DiscordUserNotFound(system)
                    
                if response.status_code != 200: # catch-all
                    raise HTTPError(response.status_code)

                system_ = System.from_json(response.json())
                url = f"{SERVER}/s/{system_.id}/switches"

            await self._respect_rate_limit()
            response = await session.get(url)

            if response.status_code == 401:
                raise AuthorizationError()

            if response.status_code == 403:
                raise AccessForbidden()
            
            if response.status_code == 404:
                if isinstance(system, System):
                    raise SystemNotFound(system.id)
                raise SystemNotFound(system)
                
            if response.status_code != 200: # catch-all
                raise HTTPError(response.status_code)

            resp = response.json()
            for item in resp:
                switch = Switch.from_json(item)
                yield switch

    def new_switch(self, members) -> Union[None, Coroutine[Any,Any,None]]:
        """Creates a new switch.
        
        Note:
            The system's `authorization token`_ must be set in order to use this method.
        
        Args:
            members (Sequence[Union[Member,str]]): A list of members that will be present in the
                new switch.

        .. _`authorization token`: https://pluralkit.me/api/#authentication
        """
        awaitable = self._new_switch(members)
        if self.async_mode:
            return awaitable
        else:
            loop = asyncio.get_event_loop()
            result = loop.run_until_complete(awaitable)
            return result

    async def _new_switch(self, members: List[Union[str, Member]]) -> None:
        if self.token is None:
            raise AuthorizationError()
        
        url = f"{SERVER}/s/switches"
        
        members = [m.id if type(m) is Member else m for m in members]

        payload = json.dumps({"members": members}, ensure_ascii=False)
        
        async with httpx.AsyncClient(headers=self.content_headers) as session:
            
            await self._respect_rate_limit()
            response = await session.post(url, data=payload)

            if response.status_code == 401:
                raise AuthorizationError()
            
            if response.status_code == 403:
                raise AccessForbidden()
                
            if response.status_code != 204: # catch-all
                raise HTTPError(response.status_code)

        return None
    
    def get_message(self, message: Union[str, int, Message]) \
    -> Union[Message, Coroutine[Any,Any,Message]]:
        """Fetches a message proxied by pluralkit
        
        Note:
            Messages proxied by pluralkit can be fetched either with the proxy message's id, or the
            id of the original message that triggered the proxy.
        
        Args:
            Message: The message to be fetched. Can be a Discord message ID (as a string or
                integer) or a Message object.

        Returns:
            Message: The message object.
        """
        awaitable = self._get_message(message)
        if self.async_mode:
            return awaitable
        else:
            loop = asyncio.get_event_loop()
            result = loop.run_until_complete(awaitable)
            return result
        
    async def _get_message(self, message: Union[str, int, Message]) -> Message:
        if isinstance(message, (str, int)):
            url = f"{SERVER}/msg/{message}"
        elif isinstance(message, Message):
            url = f"{SERVER}/msg/{message.id}"

        async with httpx.AsyncClient(headers=self.headers) as session:

            await self._respect_rate_limit()
            response = await session.get(url)

            if response.status_code == 401:
                raise AuthorizationError()
            
            if response.status_code == 403:
                raise AccessForbidden()
            
            if response.status_code != 200: # catch-all
                raise HTTPError(response.status_code)

            resp = response.json()
            return Message.from_json(resp)



import os,sys, pathlib
currentdir = pathlib.Path(__file__).parent
parentdir = os.path.dirname(currentdir)
sys.path.insert(0,parentdir) 

from pluralkit import Client, System

token = os.environ['TOKEN']
pk = Client(token, async_mode=False)

def test_get_system():
    expected_system_info = {"id":"exmpl",
                             "name":"PluralKit Example System",
                             "description":"This is an example of what a PluralKit system card "
                             "looks like. This description can be up to 1000 characters long, and "
                             "can contain **formatting** and 💎 emojis ✨. What you see above, "
                             "\"PluralKit Example System\", is the system's name. \n \nThe system's" 
                             " *tag*, displayed above, is added to the end of the member name "
                             "when proxying. This helps identify systems cohesively. \n \nFor a "
                             "list of commands change that change system cards like these, type "
                             "`pk;commands system`. \n  \nThis system has a secret private member "
                             "with the ID `cmpuv`. They won't show up in the member list, but you "
                             "can look them up to by ID see what information is visible.",
                             "tag":"| PluralKit 🦊",
                             "avatar_url":"http://placekitten.com/512/512",
                             "created":"2020-01-12T02:00:33.387824Z",
                             "tz":"UTC",
                             "description_privacy":None,
                             "member_list_privacy":None,
                             "front_privacy":None,
                             "front_history_privacy":None}
    
    expected_system = System.from_json(expected_system_info)
    
    system = pk.get_system("exmpl")
    
    assert system._deep_equal(expected_system)

def test_get_members():
    pk.get_members()

def test_edit_system():
    new_sys = {
               "id":"lmzyh",
               "name":"Testing...system",
               "description":None,
               "tag":None,
               "avatar_url":None,
               "created":"2021-06-14T01:48:47.903899Z",
               "tz":"UTC",
               "description_privacy":None,
               "member_list_privacy":None,
               "front_privacy":None,
               "front_history_privacy":None
               }
    
    new_sys = System.from_json(new_sys)
    
    pk.edit_system(new_sys)

def test_get_fronters():
    pk.get_fronters()

def test_get_switches():
    pk.get_switches()


import os,sys, pathlib
currentdir = pathlib.Path(__file__).parent
parentdir = os.path.dirname(currentdir)
sys.path.insert(0,parentdir) 

from pluralkit import Client

token = os.environ['TOKEN']
pk = Client(token, async_mode=False)

def test_new_switch():
    members = ["daoof", "cewel", "irqkk", "xadsg"]
    fronters = pk.get_fronters()
    for member in members:
        if not [member] == fronters:
            pk.new_switch([member])

def test_get_message():
    pk.get_message(860690364317958164)


from .client import Client
from .models import (
    Member,
    System,
    ProxyTag,
    ProxyTags,
    Switch,
    Privacy,
    Timestamp,
    Color,
    Birthday,
    Timezone,
    Message,
)
from . import errors




from . import v1
from . import v2

from .v2.__init__ import *

from .__version__ import version_info, __version__

__title__ = "pluralkit"
__author__ = "Madison Landry, Alyx Warner"
__copyright__ = "Copyright 2021-present Madison Landry, Alyx Warner"



 
import os,sys, pathlib
currentdir = pathlib.Path(__file__).parent
parentdir = os.path.dirname(currentdir)
sys.path.insert(0,parentdir) 

from pluralkit import Client, Member

token = os.environ['TOKEN']
pk = Client(token, async_mode=False)
    
def test_get_member():

    
    reference_member_info = {
              "id":"gaznz",
              "name":"Tester T. Testington",
              "color": None,
              "display_name": None,
              "birthday": None,
              "pronouns": None,
              "avatar_url": None,
              "description":"Here is an example of another member, albeit with less properties set." 
              " It's possible to have members without proxy tags - useful for record-keeping.",
              "proxy_tags":[],
              "keep_proxy": False,
              "privacy":None,
              "visibility":None,
              "name_privacy":None,
              "description_privacy":None,
              "birthday_privacy":None,
              "pronoun_privacy":None,
              "avatar_privacy":None,
              "metadata_privacy":None,
              "created":"2020-01-12T02:21:26.274746Z"
              }
    
    reference_member = Member.from_json(reference_member_info)
    
    test_member = pk.get_member("gaznz")
        
    try:
        assert test_member._deep_equal(other=reference_member) is True
    except AssertionError:
        for key, value in test_member.__dict__.items():
            if key not in ("id", "created"):
                try:
                    assert value == reference_member.__dict__[key]
                except AssertionError as e:
                    print(f"Error occured at {key}: {value}")
                    assert value == reference_member.__dict__[key]

def test_edit_member():
    
    choices = {
        "id": ["cewel"],
        "name": ["test_1", "test_2", "test_3"],
        "display_name": ["testing_1", "testing_2", "testing_3"],
        "description": ["testing..._1", "testing..._2", "testing..._3"],
        "pronouns": ["they/them", "she/her", "he/him"],
        "color": ["red", "#ffffff", ""],
        "created": ["2021-06-14T01:48:54.911915Z"],
        "avatar_url": ["https://cdn.pixabay.com/photo/2015/04/23/22/00/tree-736885__480.jpg", 
                       "https://image.shutterstock.com/image-photo/mountains-under-mist-morning-"
                       "amazing-260nw-1725825019.jpg"],
        "birthday": ["1904-02-28", "2025-06-12", "0834-12-23"],
        "proxy_tags": [[{"prefix": "t;"}], [{"suffix": ";t"}], [{"prefix": "te;"}]],
        "keep_proxy": [True, False],
        "visibility": ["public", "private"],
        "name_privacy": ["public", "private"],
        "description_privacy": ["public", "private"],
        "avatar_privacy": ["public", "private"],
        "birthday_privacy": ["public", "private"],
        "pronoun_privacy": ["public", "private"],
        "metadata_privacy": ["public", "private"],
    }

    
    from random import choice
    
    member = {}
    for key, value in choices.items():
        i = choice(value)
        member[key] = i
        
    m = Member.from_json(member)
    member = pk.edit_member(member_id=m.id, name=m.name, display_name=m.display_name, 
                            description=m.description, pronouns=m.pronouns, color=m.color, 
                            avatar_url=m.avatar_url, birthday=m.birthday, proxy_tags=m.proxy_tags, 
                            keep_proxy=m.keep_proxy, visibility=m.visibility, 
                            name_privacy=m.name_privacy, description_privacy=m.description_privacy, 
                            avatar_privacy=m.avatar_privacy, birthday_privacy=m.birthday_privacy, 
                            pronoun_privacy=m.pronoun_privacy, metadata_privacy=m.metadata_privacy)
        
    try:
        assert member._deep_equal(other=m, new_member=True) is True
    except AssertionError:
        for key, value in member.__dict__.items():
            if key not in ("id", "created"):
                try:
                    assert value == m.__dict__[key]
                except AssertionError as e:
                    print(f"Error occured at {key}: {value}")
                    assert value == m.__dict__[key]

def test_new_member():
    
    new_member_info = {
              "id":"gaznz",
              "name":"Tester T. Testington",
              "color": None,
              "display_name": None,
              "birthday": None,
              "pronouns": None,
              "avatar_url": None,
              "description":"Here is an example of another member, albeit with less properties set." 
              " It's possible to have members without proxy tags - useful for record-keeping.",
              "proxy_tags":[],
              "keep_proxy": False,
              "privacy":None,
              "visibility":None,
              "name_privacy":None,
              "description_privacy":None,
              "birthday_privacy":None,
              "pronoun_privacy":None,
              "avatar_privacy":None,
              "metadata_privacy":None,
              "created":"2020-01-12T02:21:26.274746Z"
              }
    new_member = Member.from_json(new_member_info)
    
    member = pk.new_member(name=new_member.name, color=new_member.color, 
                           display_name=new_member.display_name, 
                           birthday=new_member.birthday, pronouns=new_member.pronouns, 
                           avatar_url=new_member.avatar_url, description=new_member.description, 
                           visibility=new_member.visibility, name_privacy=new_member.name_privacy, 
                           birthday_privacy=new_member.birthday_privacy, 
                           keep_proxy=new_member.keep_proxy, proxy_tags=new_member.proxy_tags, 
                           pronoun_privacy=new_member.pronoun_privacy, 
                           description_privacy=new_member.description_privacy, 
                           metadata_privacy=new_member.metadata_privacy, 
                           avatar_privacy=new_member.avatar_privacy)
    try:
        assert member._deep_equal(other=new_member, new_member=True) is True
    except AssertionError:
        for key, value in member.__dict__.items():
            if key not in ("id", "created"):
                try:
                    assert value == new_member.__dict__[key]
                except AssertionError:
                    try:
                        pk.delete_member(member.id)
                    except:
                        print(f"Unable to delete member {member.id}")
                    print(f"Error occured at {key}: {value}")
                    assert value == new_member.__dict__[key]
            
    pk.delete_member(member.id)
        



from http.client import responses as RESPONSE_CODES

__all__ = (
    "PluralKitException",
    "AuthorizationError",
    "SystemNotFound",
    "MemberNotFound",
    "DiscordUserNotFound",
    "AccessForbidden",
    "InvalidKwarg",
    "InvalidBirthday",
    "HTTPError"
)

class PluralKitException(Exception):
    """Base exception class for pk.py
    """
    ...

class AuthorizationError(PluralKitException):
    """Thrown when the authorization token passed to PluralKit's API is invalid (or missing).
    """
    def __init__(self):
        super().__init__((
            "System token seems to be missing or invalid. Can you check that you entered it in "
            "correctly?"
        ))

class SystemNotFound(PluralKitException):
    """Thrown when the system ID is apparently not in PluralKit's database.
    """
    def __init__(self, id):
        super().__init__(
            f"System with the given ID ({id}) was not found."
        )

class MemberNotFound(PluralKitException):
    """Thrown when the member ID is apparently not in PluralKit's database.
    """
    def __init__(self, id):
        super().__init__(
            f"Member with the given ID ({id}) was not found."
        )

class DiscordUserNotFound(PluralKitException):
    """Thrown when the Discord user ID is apparently not associated with a PluralKit system.
    """
    def __init__(self, id):
        super().__init__(
            f"The given Discord user ID ({id}) does not seem to correspond to any systems."
        )

class AccessForbidden(PluralKitException):
    """Thrown when attempting to access private info.
    """
    def __init__(self):
        super().__init__((
            "The system's requested info is private and the client authorization token is either "
            "missing, invalid, or does not correspond to the requested system."
        ))

class InvalidKwarg(PluralKitException):
    """
    Thrown when an invalid field is passed in a POST or PATCH request
    """
    def __init__(self, key):
        super().__init__(
            f"A keyworded argument was passed that will not be accepted by the server: `{key}`"
        )

class InvalidBirthday(PluralKitException):
    """
    Thrown when an invalid string is passed for the "Birthday" field of a member object. 
    (Must be yyyy-mm-dd)
    """
    def __init__(self, string):
        super().__init__(
            f"`{string}` is not a valid yyyy-mm-dd date or datetime.datetime object"
        )

class HTTPError(PluralKitException):
    """Thrown when pluralkit.py receives an unexpected status code.
    """
    def __init__(self, status_code):
        super().__init__(
            f"Received unexpected HTTP code '{status_code} {RESPONSE_CODES[status_code]}'. " \
            f"Check your connection or whether PluralKit's API is up."
        )



from typing import (
    Any,
    Union, Optional,
    Tuple, List, Set, Sequence, Dict,
    Callable,
    AsyncGenerator, Awaitable,
)
import datetime
import asyncio
from http.client import responses as RESPONSE_CODES
from functools import wraps

import httpx

from .models import (
    Model,
    AutoproxyMode,
    MemberId, SystemId, GroupId, SwitchId,
    Member, System, Group, Switch, Message,
    MemberGuildSettings, SystemGuildSettings,
    SystemSettings, AutoproxySettings,
    Timestamp, Privacy,
    _PATCHABLE_SYSTEM_KEYS,
    _PATCHABLE_MEMBER_KEYS,
    _PATCHABLE_GROUP_KEYS,
    _PATCHABLE_SWITCH_KEYS,
    _PATCHABLE_AUTOPROXY_SETTINGS_KEYS,
    _PATCHABLE_SYSTEM_SETTINGS_KEYS,
    _PATCHABLE_SYSTEM_GUILD_SETTINGS_KEYS,
    _PATCHABLE_MEMBER_GUILD_SETTINGS_KEYS,
    _PRIVACY_ASSOCIATED_KEYS,
)
from .errors import (
    PluralKitException,
    HTTPError,
    GenericBadRequest,
    NotFound,
    SystemNotFound,
    MemberNotFound,
    GroupNotFound,
    SwitchNotFound,
    MessageNotFound,
    GuildNotFound,
    Unauthorized,
    NotOwnSystem,
    NotOwnMember,
    NotOwnGroup,
    GENERIC_ERROR_CODE_LOOKUP,
    SYSTEM_ERROR_CODE_LOOKUP,
    MEMBER_ERROR_CODE_LOOKUP,
    GROUP_ERROR_CODE_LOOKUP,
    MESSAGE_ERROR_CODE_LOOKUP,
    SWITCH_ERROR_CODE_LOOKUP,
    GUILD_ERROR_CODE_LOOKUP,
)

SERVER = "https://api.pluralkit.me/v2"

async def aiter(generator: Awaitable) -> AsyncGenerator:
    """For conversion of any awaitables to async generators/sequences

    Applies to async_mode=True Sequence methods
    """
    items = await generator
    for item in items:
        yield item
        await asyncio.sleep(0)

class Client:
    """Represents a client that interacts with the PluralKit API.

    Args:
        token: The PluralKit `authorization token`_, from the ``pk;token`` command.

    Keyword args:
        async_mode: Whether the client runs asynchronously (``True``, default) or not (``False``).
        user_agent: The User-Agent header to use with the API.
        loop: The `asyncio` event loop to use (if ``async_mode=True``), default is the current
            event loop.

    Attributes:
        token: The client's PluralKit authorization token.
        async_mode: Whether the client runs asynchronously (``True``) or not (``False``).
        headers: The headers the client uses to communicate with the API.

    .. _`authorization token`: https://pluralkit.me/api/#authentication
    """
    def __init__(self, token: Optional[str]=None, *,
        async_mode: bool=True,
        user_agent: Optional[str]=None,
        loop: asyncio.AbstractEventLoop=None,
    ):
        # core factors
        self.async_mode = async_mode
        self.loop = loop
        self.id = None
        self._token = token

        # initialize rate limit handling
        self._rate_limit = 2 # default is 2 requests per second
        self._rate_limit_remaining = 0
        self._rate_limit_reset_time = datetime.datetime.now()

        # set up headers
        self.headers = {}
        if user_agent: self.headers["User-Agent"] = user_agent
        if token: self.headers["Authorization"] = token

    @property
    def token(self):
        return self._token

    @token.setter
    def token(self, new_token):
        self._token = new_token
        if new_token is None:
            del self.headers["Authorization"]
        else:
            self.headers["Authorization"] = new_token

    # ==========================
    #  abstract request methods
    # ==========================

    async def _respect_rate_limit(self):
        """Respects the rate limit by waiting if necessary."""
        if self._rate_limit_remaining == 0:
            now = datetime.datetime.now()
            if self._rate_limit_reset_time < now:
                self._rate_limit_remaining = self._rate_limit
            else:
                await asyncio.sleep((self._rate_limit_reset_time - now).total_seconds())
                self._rate_limit_remaining = self._rate_limit
            self._rate_limit_reset_time = now + datetime.timedelta(seconds=1) # until otherwise specified
 
    def _update_rate_limits(self, headers):
        """Updates the rate limits based on the returned headers."""
        if "X-RateLimit-Limit" in headers:
            self._rate_limit = int(headers["X-RateLimit-Limit"])
        if "X-RateLimit-Remaining" in headers:
            self._rate_limit_remaining = int(headers["X-RateLimit-Remaining"])
        if "X-RateLimit-Reset" in headers:
            timestamp = float(headers["X-RateLimit-Reset"]) / 1000.0
            self._rate_limit_reset_time = datetime.datetime.fromtimestamp(timestamp)
            #print("reset at:", self._rate_limit_reset_time - datetime.datetime.now())

    async def _request_something(self,
        # required positional arguments
        kind: str,
        url_template: str,
        ModelConstructor,
        expected_code: int,
        error_lookups: dict,
        *,
        # reference IDs
        system: Union[SystemId,int,None]=None, # (int too, because it can be a Discord user ID)
        member: Optional[MemberId]=None,
        group: Optional[GroupId]=None,
        switch: Optional[SwitchId]=None,
        message: Optional[int]=None,
        guild_id: Optional[int]=None,
        # optional payloads
        payload: Optional[dict]=None,
        params: Optional[dict]=None, # for query strings
    ):
        # put together url from given arguments
        pieces = {"SERVER": SERVER}

        if "system_ref" in url_template:
            if system is None:
                if not self.token: raise Unauthorized() # pass in your token please
                pieces["system_ref"] = "@me"
            else:
                pieces["system_ref"] = str(system)

        map_ = {
            "member_ref": member,
            "group_ref": group,
            "switch_ref": switch,
            "message_ref": message,
            "guild_id": guild_id,
        }
        for ref_name, arg in map_.items():
            if ref_name in url_template:
                pieces[ref_name] = str(arg)

        url = url_template.format(**pieces)

        # respect rate limit
        await self._respect_rate_limit()

        # make the request
        async with httpx.AsyncClient(headers=self.headers) as session:
            request_func = getattr(session, kind.lower())
            kwargs = {}
            if params is not None: kwargs["params"] = params
            if payload is not None: kwargs["json"] = payload
            response = await request_func(url, **kwargs)

            # update rate limit mechanics
            headers = response.headers
            self._update_rate_limits(headers)

            # analyze returned info
            code = response.status_code
            returned = response.json() if response.text else ""
            if code != expected_code:
                if code in error_lookups:
                    if isinstance(returned, dict) and "message" in returned:
                        msg = "{code}: {message}".format(**returned)
                    else:
                        msg = f"{code}: {response.text!r}" if response.text else f"{code}"
                    error = error_lookups[code](msg)
                else:
                    error = HTTPError(code)
                raise error

            # convert received json to return type
            #print(returned)
            converted = ModelConstructor(returned)

        # return
        return converted

    # streamline the async_mode=True vs. False difference
    def _async_mode_handler(wrapped_function):
        @wraps(wrapped_function)
        def wrapped(instance, *args, **kwargs):
            awaitable = wrapped_function(instance, *args, **kwargs)
            if instance.async_mode:
                return_type = wrapped_function.__annotations__.get("return")
                try:
                    is_generator = return_type is not None and return_type._name == Sequence._name
                except AttributeError:
                    is_generator = False
                return aiter(awaitable) if is_generator else awaitable
            else:
                loop = instance.loop if instance.loop is not None else asyncio.get_event_loop()
                result = loop.run_until_complete(awaitable)
                return result
        return wrapped

    # ==============
    #  Main methods
    # ==============

    # 
    #  get
    # 

    @_async_mode_handler
    def get_system(self, system: Union[SystemId,int,None]=None) -> System:
        """Get a system by its reference ID.

        Args:
            system: System reference; a system's ID (`SystemId`) or the ID of a Discord account
                linked to the system. Default is ``None``, for the system corresponding to the
                client's authorization token.
        """
        if isinstance(system, System): system = system.id
        return self._request_something(
            "GET",
            "{SERVER}/systems/{system_ref}",
            System,
            200,
            SYSTEM_ERROR_CODE_LOOKUP,
            system=system,
        )

    @_async_mode_handler
    def get_member(self, member: Union[MemberId,str]) -> Member:
        """Get a member by its reference ID.

        Args:
            member: Member reference (`MemberId`).
        """
        return self._request_something(
            "GET",
            "{SERVER}/members/{member_ref}",
            Member,
            200,
            MEMBER_ERROR_CODE_LOOKUP,
            member=member,
        )

    @_async_mode_handler
    def get_group(self, group: Union[GroupId,str]) -> Group:
        """Get a group by its reference ID.

        Args:
            group: Group reference (`GroupId`).
        """
        return self._request_something(
            "GET",
            "{SERVER}/groups/{group_ref}",
            Group,
            200,
            GROUP_ERROR_CODE_LOOKUP,
            group=group,
        )

    @_async_mode_handler
    def get_message(self, message: int) -> Message:
        """Get information about a proxied message.

        Args:
            message: The Discord ID of a proxied message, or the Discord ID of the message that
                sent the proxy.
        """
        return self._request_something(
            "GET",
            "{SERVER}/messages/{message_ref}",
            Message,
            200,
            MESSAGE_ERROR_CODE_LOOKUP,
            message=message,
        )

    @_async_mode_handler
    def get_switches(self, system: Union[SystemId,int,None]=None, *,
        before: Timestamp=None,
        limit: int=None,
    ) -> Sequence[Switch]:
        """Get list of system switches.

        Returns at most 100 switches. To get additional, specify the ``before`` parameter.

        Note:
            Because this is a batch API call, the `Switch` objects returned by this method have
            `MemberId` objects for `Switch.members` rather than full `Member` objects.

        Hint:
            This method returns an async generator when the `Client` is in asynchronous mode
            (and with ``async_mode=False``, it returns a list), so use the ``async for`` syntax instead of
            ``await``: ::

                async for switch in client.get_switches(...):
                    ...

            If you find a list preferable to an async generator, use a list comprehension such
            as ::

                switches = [s async for s in client.get_switches(...)]
                ...

        Args:
            system: System reference; a system's ID (`SystemId`) or the ID of a Discord account
                linked to the system. Default is ``None``, for the system corresponding to the
                client's authoirzation token.

        Keyword args:
            before: Timestamp before which to get latest switches from. Default is ``None``
                ("now").
            limit: Number of switches to return. Default (and maximum) is 100.
        """
        params = {}
        if before is not None: params["before"] = before.json()
        if limit is not None: params["limit"] = limit
        if not params: params = None
        return self._request_something(
            "GET",
            "{SERVER}/systems/{system_ref}/switches",
            lambda list_: [Switch(item) for item in list_],
            200,
            GENERIC_ERROR_CODE_LOOKUP,
            system=system,
            params=params,
        )

    @_async_mode_handler
    def get_fronters(self, system: Union[SystemId,int,None]=None) \
    -> Sequence[Member]:
        """Get list of current fronters.

        Hint:
            This method returns an async generator when the `Client` is in asynchronous mode
            (and with ``async_mode=False``, it returns a list), so use the ``async for`` syntax instead of
            ``await``: ::

                async for member in client.get_fronters(...):
                    ...

            If you find a list preferable to an async generator, use a list comprehension such
            as ::

                fronters = [m async for m in client.get_fronters(...)]
                ...

        Args:
            system: System reference; a system's ID (`SystemId`) or the ID of a Discord account
                linked to the system. Default is ``None``, for the system corresponding to the
                client's authoirzation token.
        """
        return self._request_something(
            "GET",
            "{SERVER}/systems/{system_ref}/fronters",
            lambda json: [Member(item) for item in json["members"]],
            200,
            GENERIC_ERROR_CODE_LOOKUP,
            system=system,
        )

    @_async_mode_handler
    def get_switch(self, switch: Union[SwitchId,str], system: Union[SystemId,int,None]=None
    ) -> Switch:
        """Get switch information by its reference ID.

        Args:
            switch: Switch reference (`SwitchId`).
        """
        return self._request_something(
            "GET",
            "{SERVER}/systems/{system_ref}/switches/{switch_ref}",
            Switch,
            200,
            SWITCH_ERROR_CODE_LOOKUP,
            system=system,
            switch=switch,
        )

    @_async_mode_handler
    def get_members(self, system: Union[SystemId,int,None]=None) -> Sequence[Member]:
        """Get the list of a system's members.

        Hint:
            This method returns an async generator when the `Client` is in asynchronous mode
            (and with ``async_mode=False``, it returns a list), so use the ``async for`` syntax instead of
            ``await``. For example: ::

                async for member in client.get_members(...):
                    ...

            If you find a list preferable to an async generator, use a list comprehension such
            as ::

                members = [m async for m in client.get_members(...)]
                ...

        Args:
            system: System reference; a system's ID (`SystemId`) or the ID of a Discord account
                linked to the system. Default is ``None``, for the system corresponding to the
                client's authoirzation token.
        """
        return self._request_something(
            "GET",
            "{SERVER}/systems/{system_ref}/members",
            lambda list_: [Member(m) for m in list_],
            200,
            SYSTEM_ERROR_CODE_LOOKUP,
            system=system,
        )

    @_async_mode_handler
    def get_member_groups(self, member: Union[Member,str]) -> Sequence[Group]:
        """Get the list of groups a member is a part of.

        Hint:
            This method returns an async generator when the `Client` is in asynchronous mode
            (and with ``async_mode=False``, it returns a list), so use the ``async for`` syntax instead of
            ``await``. For example: ::

                async for group in client.get_member_groups(...):
                    ...

            If you find a list preferable to an async generator, use a list comprehension such
            as ::

                groups = [g async for g in client.get_member_groups(...)]
                ...

        Args:
            member: Member reference (`MemberId`).
        """
        return self._request_something(
            "GET",
            "{SERVER}/members/{member_ref}/groups",
            lambda list_: [Group(g) for g in list_],
            200,
            MEMBER_ERROR_CODE_LOOKUP,
            member=member,
        )

    @_async_mode_handler
    def get_groups(self, system: Union[SystemId,int,None]=None) -> Sequence[Group]:
        """Get the list of a system's groups.

        Hint:
            This method returns an async generator when the `Client` is in asynchronous mode
            (and with ``async_mode=False``, it returns a list), so use the ``async for`` syntax instead of
            ``await``. For example: ::

                async for group in client.get_groups(...):
                    ...

            If you find a list preferable to an async generator, use a list comprehension such
            as ::

                groups = [g async for g in client.get_groups(...)]
                ...

        Args:
            system: System reference; a system's ID (`SystemId`) or the ID of a Discord account
                linked to the system. Default is ``None``, for the system corresponding to the
                client's authoirzation token.
        """
        return self._request_something(
            "GET",
            "{SERVER}/systems/{system_ref}/groups",
            lambda list_: [Group(g) for g in list_],
            200,
            SYSTEM_ERROR_CODE_LOOKUP,
            system=system,
        )

    @_async_mode_handler
    def get_group_members(self, group: Union[GroupId,str]) -> Sequence[Member]:
        """Get the list of members of a group.

        Hint:
            This method returns an async generator when the `Client` is in asynchronous mode
            (and with ``async_mode=False``, it returns a list), so use the ``async for`` syntax instead of
            ``await``. For example: ::

                async for member in client.get_group_members(...):
                    ...

            If you find a list preferable to an async generator, use a list comprehension such
            as ::

                members = [m async for m in client.get_group_members(...)]
                ...

        Args:
            group: Group reference (`GroupId`).
        """
        return self._request_something(
            "GET",
            "{SERVER}/groups/{group_ref}/members",
            lambda list_: [Member(m) for m in list_],
            200,
            GROUP_ERROR_CODE_LOOKUP,
            group=group,
        )

    @_async_mode_handler
    def get_system_settings(self, system: Union[SystemId,str,int,None]=None) -> SystemSettings:
        """Get a system's settings (e.g. defaults, timezone, limits).

        Args:
            system: System reference; a system's ID (`SystemId`) or the ID of a Discord account
                linked to the system. Default is ``None``, for the system corresponding to the
                client's authoirzation token.
        """
        return self._request_something(
            "GET",
            "{SERVER}/systems/{system_ref}/settings",
            SystemSettings,
            200,
            SYSTEM_ERROR_CODE_LOOKUP,
            system=system,
        )

    @_async_mode_handler
    def get_autoproxy_settings(self, guild_id: int) -> AutoproxySettings:
        """Get the autoproxy settings of your system.

        Args:
            guild_id: Discord guild ID.
        """
        return self._request_something(
            "GET",
            "{SERVER}/systems/@me/autoproxy",
            AutoproxySettings,
            200,
            GENERIC_ERROR_CODE_LOOKUP,
            params={"guild_id": guild_id},
        )

    @_async_mode_handler
    def get_system_guild_settings(self, guild_id: int, system: Union[SystemId,int,None]=None) \
    -> SystemGuildSettings:
        """Get the guild settings of a system.

        Args:
            guild_id: Discord guild ID.
            system: System reference; a system's ID (`SystemId`) or the ID of a Discord account
                linked to the system. Default is ``None``, for the system corresponding to the
                client's authoirzation token.
        """
        error_lookups = GENERIC_ERROR_CODE_LOOKUP | {403: NotOwnSystem}
        return self._request_something(
            "GET",
            "{SERVER}/systems/{system_ref}/guilds/{guild_id}",
            SystemGuildSettings,
            200,
            error_lookups,
            guild_id=guild_id,
            system=system,
        )

    @_async_mode_handler
    def get_member_guild_settings(self, guild_id: int, member: Union[MemberId,str]) \
    -> MemberGuildSettings:
        """Get the guild settings of a member.

        Args:
            guild_id: Discord guild ID.
            member: Member reference (`MemberId`).
        """
        error_lookups = GENERIC_ERROR_CODE_LOOKUP | {403: NotOwnMember}
        return self._request_something(
            "GET",
            "{SERVER}/members/{member_ref}/guilds/{guild_id}",
            MemberGuildSettings,
            200,
            error_lookups,
            member=member,
            guild_id=guild_id,
        )

    # 
    #  delete
    # 

    @_async_mode_handler
    def delete_member(self, member: Union[MemberId,str]) -> None:
        """Remove a member from your system.

        Args:
            member: Member reference (`MemberId`).
        """
        return self._request_something(
            "DELETE",
            "{SERVER}/members/{member_ref}",
            lambda x: None,
            204,
            MEMBER_ERROR_CODE_LOOKUP,
            member=member,
        )

    @_async_mode_handler
    def delete_group(self, group: Union[GroupId,Group]) -> None:
        """Remove a group from your system.

        Args:
            group: Group reference (`GroupId`).
        """
        if isinstance(group, Group): group = group.id
        return self._request_something(
            "DELETE",
            "{SERVER}/groups/{group_ref}",
            lambda x: None,
            204,
            GROUP_ERROR_CODE_LOOKUP,
            group=group,
        )

    @_async_mode_handler
    def delete_switch(self, switch: Union[SwitchId,Switch]) -> None:
        """Remove a logged switch from your system.

        Args:
            switch: Switch reference (`SwitchId`).
        """
        if isinstance(switch, Switch): switch = switch.id
        return self._request_something(
            "DELETE",
            "{SERVER}/systems/@me/switches/{switch_ref}",
            lambda x: None,
            204,
            SWITCH_ERROR_CODE_LOOKUP,
            switch=switch,
        )

    # 
    #  add, remove, set
    # 

    def _gather_args(self, args, context) -> Sequence[str]:
        ALLOWED = {
            "groups": {
                Group: lambda g: g.id.uuid,
                GroupId: lambda gid: gid.uuid,
            },
            "members": {
                Member: lambda m: m.id.uuid,
                MemberId: lambda mid: mid.uuid,
            }
        }
        arg_list = []
        for arg in args:
            for allowed_type in ALLOWED[context]:
                if isinstance(arg, allowed_type):
                    func = ALLOWED[context][allowed_type]
                    arg_list.append(func(arg))
                    break
            else:
                allowed_types = ", ".join([c.__name__ for c in ALLOWED[context].keys()])
                msg = (
                    f"{context} must be of any of these types: {allowed_types}. "
                    f"Encountered type {type(group)!r} ({group!r})."
                )
                raise ValueError(msg)
        return arg_list

    @_async_mode_handler
    def add_member_groups(self, member: Union[MemberId,str], groups) -> None:
        """Add a member to each group in a given list of groups.

        Not to be confused with `Client.add_group_members`.

        Args:
            member: The member to add to the groups.
            groups (Sequence[Union[GroupId,Group]]): The groups to add the member to.
        """
        group_list = self._gather_args(groups, "groups")
        return self._request_something(
            "POST",
            "{SERVER}/members/{member_ref}/groups/add",
            lambda x: None,
            204,
            MEMBER_ERROR_CODE_LOOKUP,
            member=member,
            payload=group_list,
        )

    @_async_mode_handler
    def remove_member_groups(self, member: Union[MemberId,str], groups) -> None:
        """Remove a member from each group in a given list of groups.

        Not to be confused with `Client.remove_group_members`.

        Tip:
            If you want to remove *all* groups from a member, consider using
            `Client.set_member_groups` with no groups.

        Args:
            member: The member to remove from the groups.
            groups (Sequence[Union[GroupId,Group]]): The groups to remove the member from.
        """
        group_list = self._gather_args(groups, "groups")
        return self._request_something(
            "POST",
            "{SERVER}/members/{member_ref}/groups/remove",
            lambda x: None,
            204,
            MEMBER_ERROR_CODE_LOOKUP,
            member=member,
            payload=group_list,
        )

    @_async_mode_handler
    def set_member_groups(self, member: Union[MemberId,str], groups) -> None:
        """Explicitly set which groups a member belongs to.

        Not to be confused with `Client.set_group_members`.

        An empty list is accepted, effectively removing the member from any & all groups.

        Args:
            member: The member for which to set the groups.
            groups (Sequence[Union[GroupId,Group]]): The groups to assign to the member.
        """
        group_list = self._gather_args(groups, "groups")
        return self._request_something(
            "POST",
            "{SERVER}/members/{member_ref}/groups/overwrite",
            lambda x: None,
            204,
            MEMBER_ERROR_CODE_LOOKUP,
            member=member,
            payload=group_list,
        )

    @_async_mode_handler
    def add_group_members(self, group: Union[GroupId,Group], members) -> None:
        """Add members to a group.

        Not to be confused with `Client.add_member_groups`.

        Args:
            group: The group to add the members to.
            members (Sequence[Union[Member,MemberId]]): The members to add to the group.
        """
        member_list = self._gather_args(members, "members")
        return self._request_something(
            "POST",
            "{SERVER}/groups/{group_ref}/members/add",
            lambda x: None,
            204,
            GROUP_ERROR_CODE_LOOKUP,
            group=group,
            payload=member_list,
        )

    @_async_mode_handler
    def remove_group_members(self, group: Union[GroupId,Group], members) -> None:
        """Remove members from a group.

        Not to be confused with `Client.remove_member_groups`.

        Tip:
            If you want to remove *all* members from a group, consider using
            `Client.set_group_members` with no members.

        Args:
            group: The group to remove the members from.
            members (Sequence[Union[Member,MemberId]]): The members to remove from the group.
        """
        member_list = self._gather_args(members, "members")
        return self._request_something(
            "POST",
            "{SERVER}/groups/{group_ref}/members/remove",
            lambda x: None,
            204,
            GROUP_ERROR_CODE_LOOKUP,
            group=group,
            payload=member_list,
        )

    @_async_mode_handler
    def set_group_members(self, group: Union[GroupId,Group], members) -> None:
        """Remove members from a group.

        Not to be confused with `Client.set_member_groups`.

        An empty list is accepted, effectively removing all members from a group.

        Args:
            group: The group to remove the members from.
            members (Sequence[Union[Member,MemberId]]): The members to assign to the group.
        """
        member_list = self._gather_args(members, "members")
        return self._request_something(
            "POST",
            "{SERVER}/groups/{group_ref}/members/overwrite",
            lambda x: None,
            204,
            GROUP_ERROR_CODE_LOOKUP,
            group=group,
            payload=member_list,
        )

    # 
    #  update (patches)
    # 

    def _check_update_keys(self, context, kwargs, allowed, require_at_least_one_arg=True):
        if require_at_least_one_arg and not kwargs:
            msg = f"{context} expects at least 1 keyword argument"
            raise TypeError(msg)
        for key in kwargs:
            if key not in allowed:
                msg = f"{context} got an unexpected keyword argument {key!r}"
                raise TypeError(msg)

    @_async_mode_handler
    def update_system(self, **kwargs) -> System:
        """Update your system.

        Keyword args:
            name (str): System name (100 character limit).
            description (Optional[str]): System description (1,000 character limit). ``None`` to
                clear.
            tag (str): System tag (79 character limit).
            pronouns (Optional[str]): System pronouns (100 character limit). ``None`` to clear.
            avatar_url (Optional[str]): System avatar url (256 character limite). ``None`` to
                clear.
            banner (Optional[str]): System banner url (256 character limite). ``None`` to clear.
            color (Color): System color.
            description_privacy (Optional[Privacy]): System description privacy. ``None`` to reset
                (to public).
            pronoun_privacy (Optional[Privacy]): System pronoun privacy. ``None`` to reset (to
                public).
            member_list_privacy (Optional[Privacy]): System member list privacy. ``None`` to reset
                (to public).
            group_list_privacy (Optional[Privacy]): System group list privacy. ``None`` to reset
                (to public).
            front_privacy (Optional[Privacy]): System current fronter privacy. ``None`` to reset
                (to public).
            front_history_privacy (Optional[Privacy]): System front history privacy. ``None`` to
                reset (to public).
        """
        self._check_update_keys("update_system()", kwargs, _PATCHABLE_SYSTEM_KEYS)
        payload = {}
        privacies = {}
        for key, value in kwargs.items():
            json_value = _PATCHABLE_SYSTEM_KEYS[key](value)
            if key in _PRIVACY_ASSOCIATED_KEYS:
                privacies[key] = json_value
            else:
                payload[key] = json_value
        if privacies: payload["privacy"] = privacies
        return self._request_something(
            "PATCH",
            "{SERVER}/systems/@me",
            System,
            200,
            SYSTEM_ERROR_CODE_LOOKUP,
            payload=payload,
        )

    @_async_mode_handler
    def update_system_settings(self, **kwargs) -> SystemSettings:
        """Update your system settings.

        Keyword args:
            timezone (Timezone): System timezone. ``None`` resets to UTC.
            pings_enabled (bool): Whether other users are able to ping your system (with a 🔔
                reaction).
            latch_timeout (Optional[int]): For autoproxy latch mode, this sets the amount of time
                elapsed since the last proxied message in the server that resets the latch.
                ``None`` resets to default (6 hr).
            member_default_private (bool): Whether member privacy is automatically set to private for new members.
                Applies only to new members created by the bot and *not* created by the PluralKit API directly.
            group_default_private (bool): Whether group privacy is automatically set to private for new groups. Applies
                only to new groups created by the bot and *not* created by the PluralKit API directly.
            show_private_info (bool): Whether the bot shows the system's own private information without a ``-private``
                flag.
        """
        self._check_update_keys("update_system_settings()",
            kwargs, _PATCHABLE_SYSTEM_SETTINGS_KEYS)
        payload = {}
        for key, value in kwargs.items():
            json_value = _PATCHABLE_SYSTEM_SETTINGS_KEYS[key](value)
            payload[key] = json_value
        return self._request_something(
            "PATCH",
            "{SERVER}/systems/@me/settings",
            SystemSettings,
            200,
            SYSTEM_ERROR_CODE_LOOKUP,
            payload=payload
        )

    @_async_mode_handler
    def update_autoproxy_settings(self, guild_id: int, autoproxy_mode: AutoproxyMode, **kwargs) -> AutoproxySettings:
        """Update your system's autoproxy settings.

        Args:
            guild_id: Discord guild ID.
            autoproxy_mode (AutoproxyMode): The autoproxy mode to use for this Discord server.

        Keyword args:
            autoproxy_member (Union[MemberId,Member,None]): The member to autoproxy, when autoproxy
                mode is set to `AutoproxyMode.MEMBER`.
        """
        self._check_update_keys("update_autoproxy_settings()",
            kwargs, _PATCHABLE_AUTOPROXY_SETTINGS_KEYS, require_at_least_one_arg=False)
        autoproxy_mode = AutoproxyMode(autoproxy_mode)
        payload = {"autoproxy_mode": autoproxy_mode.json()}
        for key, value in kwargs.items():
            json_value = _PATCHABLE_AUTOPROXY_SETTINGS_KEYS[key](value)
            payload[key] = json_value
        # API quirk?
        if autoproxy_mode == AutoproxyMode.FRONT and "autoproxy_member" in kwargs:
            kwargs["autoproxy_member"] = None
        return self._request_something(
            "PATCH",
            "{SERVER}/systems/@me/autoproxy",
            AutoproxySettings,
            200,
            GUILD_ERROR_CODE_LOOKUP,
            payload=payload,
            params={"guild_id": guild_id},
        )

    @_async_mode_handler
    def update_system_guild_settings(self, guild_id: int, **kwargs) -> SystemGuildSettings:
        """Update your system's guild settings.

        Args:
            guild_id: Discord guild ID.

        Keyword args:
            proxying_enabled (bool): Whether the system can proxy messages in this guild.
            tag (Optional[str]): System tag in this guild (79 character limit). Use ``None`` to
                clear.
            tag_enabled (bool): Whether the system tag is enabled for this guild.
        """
        self._check_update_keys("update_system_guild_settings()",
            kwargs, _PATCHABLE_SYSTEM_GUILD_SETTINGS_KEYS)
        payload = {}
        for key, value in kwargs.items():
            json_value = _PATCHABLE_SYSTEM_GUILD_SETTINGS_KEYS[key](value)
            payload[key] = json_value
        return self._request_something(
            "PATCH",
            "{SERVER}/systems/@me/guilds/{guild_id}",
            SystemGuildSettings,
            200,
            GUILD_ERROR_CODE_LOOKUP,
            guild_id=guild_id,
            payload=payload
        )

    @_async_mode_handler
    def update_member(self, member: Union[MemberId,Member], **kwargs) -> Member:
        """Update a system member.

        Args:
            member: Member reference (`MemberId`).

        Keyword args:
            name (str): Member name (100 character limit).
            display_name (Optional[str]): Member display name (100 character limit). ``None`` to
                clear.
            color (Color): Member color.
            birthday (Optional[Birthday]): Member birthday. ``None`` to clear.
            pronouns (Optional[str]): Member pronouns (100 character limit). ``None`` to clear.
            avatar_url (Optional[str]): Member avatar url (256 character limit). ``None`` to clear.
            banner (Optional[str]): Member banner url (256 character limit). ``None`` to clear.
            description (Optional[str]): Member description (1,000 character limit). ``None`` to
                clear.
            proxy_tags (Union[ProxyTag,ProxyTags,Sequence[ProxyTag]]): Member proxy tags. Use an
                empty list to remove the member's set of proxy tags.
            keep_proxy (bool): Whether to keep the proxy prefix and/or suffix for this member's
                proxied messages.
            visibility (Optional[Privacy]): Whether this member is visible to others (i.e. in
                member lists). ``None`` to reset to default (public).
            name_privacy (Optional[Privacy]): Whether the member name is visible to others or only
                the display name. ``None`` to reset to default (public).
            description_privacy (Optional[Privacy]): Whether the member description is visible to
                others. ``None`` to reset to default (public).
            birthday_privacy (Optional[Privacy]): Whether the member birthday is visible to others.
                ``None`` to reset to default (public).
            pronoun_privacy (Optional[Privacy]): Whether the member pronouns are visible to others.
                ``None`` to reset to default (public).
            avatar_privacy (Optional[Privacy]): Whether the member avatar is visible to others.
                ``None`` to reset to default (public).
            metadata_privacy (Optional[Privacy]): Whether the member's metadata (i.e. created
                timestamp, message count) is visible to others. ``None`` to reset to default
                (public).
        """
        if isinstance(member, Member): member = member.id
        self._check_update_keys("update_member()", kwargs, _PATCHABLE_MEMBER_KEYS)
        payload = {}
        privacies = {}
        for key, value in kwargs.items():
            json_value = _PATCHABLE_MEMBER_KEYS[key](value)
            if key in _PRIVACY_ASSOCIATED_KEYS:
                privacies[key] = json_value
            else:
                payload[key] = json_value
        if privacies: payload["privacy"] = privacies
        return self._request_something(
            "PATCH",
            "{SERVER}/members/{member_ref}",
            Member,
            200,
            MEMBER_ERROR_CODE_LOOKUP,
            member=member,
            payload=payload,
        )

    @_async_mode_handler
    def update_member_guild_settings(self,
        guild_id: int, member: Union[Member,MemberId], **kwargs) -> MemberGuildSettings:
        """Update a member's guild settings.

        Args:
            guild_id: Discord guild ID.
            member: Member reference (`MemberId`).

        Keyword args:
            display_name (Optional[str]): Member display name (100 character limit). ``None`` to
                clear.
            avatar_url (Optional[str]): Member avatar url (256 character limit). ``None`` to clear.
        """
        self._check_update_keys("update_member_guild_settings()",
            kwargs, _PATCHABLE_MEMBER_GUILD_SETTINGS_KEYS)
        payload = {}
        for key, value in kwargs.items():
            json_value = _PATCHABLE_MEMBER_GUILD_SETTINGS_KEYS[key](value)
            payload[key] = json_value
        if isinstance(member, Member): member = member.id
        error_lookups = GENERIC_ERROR_CODE_LOOKUP | {403: NotOwnMember}
        return self._request_something(
            "PATCH",
            "{SERVER}/members/{member_ref}/guilds/{guild_id}",
            MemberGuildSettings,
            200,
            error_lookups,
            member=member,
            guild_id=guild_id,
            payload=payload
        )

    @_async_mode_handler
    def update_group(self, group: Union[GroupId,Group], **kwargs) -> Group:
        """Update a group's settings.

        Args:
            group: Group reference (`GroupId`).

        Keyword args:
            name (str): Group name (100 character limit).
            display_name (Optional[str]): Group display name (100 character limit). ``None`` to
                clear.
            description (Optional[str]): Group description (1,000 character limit). ``None`` to
                clear.
            icon (Optional[str]): Group icon url (256 character limit). ``None`` to clear.
            banner (Optional[str]): Group banner url (256 character limit). ``None`` to clear.
            color (Color): Group color.
            name_privacy (Optional[Privacy]): Whether the group name is visible to others or only
                the display name. ``None`` to reset to default (public).
            description_privacy (Optional[Privacy]): Whether the group description is visilbe to
                others. ``None`` to reset to default (public).
            icon_privacy (Optional[Privacy]): Whether the group icon is visible to others. ``None``
                to reset to default (public).
            list_privacy (Optional[Privacy]): Whether the group member list is visible to others.
                ``None`` to reset to default (public).
            metadata_privacy (Optional[Privacy]): Whether the groups's metadata (i.e. created
                timestamp) is visible to others. ``None`` to reset to default (public).
            visibility (Optional[Privacy]): Whether this group is visible to others (i.e. in group
                lists). ``None`` to reset to default (public).
        """
        self._check_update_keys("update_group()", kwargs, _PATCHABLE_GROUP_KEYS)
        payload = {}
        privacies = {}
        for key, value in kwargs.items():
            json_value = _PATCHABLE_GROUP_KEYS[key](value)
            if key in _PRIVACY_ASSOCIATED_KEYS:
                privacies[key] = json_value
            else:
                payload[key] = json_value
        if privacies: payload["privacy"] = privacies
        if isinstance(group, Group): group = group.id
        return self._request_something(
            "PATCH",
            "{SERVER}/groups/{group_ref}",
            Group,
            200,
            GROUP_ERROR_CODE_LOOKUP,
            group=group,
            payload=payload,
        )

    # note: two endpoints per this method
    @_async_mode_handler
    def update_switch(self, switch: Union[SwitchId,Switch], **kwargs) -> Switch:
        """Update a logged switch.

        Args:
            switch: Switch reference (`SwitchId`).

        Keyword args:
            timestamp: When the switch happened.
            members (Sequence[Union[Member,MemberId]]): The list of member IDs.
        """
        self._check_update_keys("update_switch()", kwargs, _PATCHABLE_SWITCH_KEYS)
        awaitables = []
        if "members" in kwargs:
            members = _PATCHABLE_SWITCH_KEYS["members"](kwargs["members"])
            awaitable = self._request_something(
                "PATCH",
                "{SERVER}/systems/@me/switches/{switch_ref}/members",
                Switch,
                200,
                SWITCH_ERROR_CODE_LOOKUP,
                switch=switch,
                payload=members,
            )
            awaitables.append(awaitable)
        if "timestamp" in kwargs:
            timestamp = _PATCHABLE_SWITCH_KEYS["timestamp"](kwargs["timestamp"])
            awaitable = self._request_something(
                "PATCH",
                "{SERVER}/systems/@me/switches/{switch_ref}",
                Switch,
                200,
                SWITCH_ERROR_CODE_LOOKUP,
                switch=switch,
                payload={"timestamp": timestamp},
            )
            awaitables.append(awaitable)
        
        async def send_request(awaitables):
            results = []
            for a in awaitables:
                results.append(await a) # purposefully block
                                        # to return only most recent Switch model
            return results[-1]

        return send_request(awaitables)

    @_async_mode_handler
    def new_member(self, name: str, **kwargs) -> Member:
        """Create a new member in your system.

        Args:
            name: The name of the new member. 100 character limit.

        Keyword args:
            display_name (str): Member display name (100 character limit).
            color (Color): Member color.
            birthday (Birthday): Member birthday.
            pronouns (str): Member pronouns (100 character limit).
            avatar_url (str): Member avatar url (256 character limit).
            banner (str): Member banner url (256 character limit).
            description (str): Member description (1,000 character limit).
            proxy_tags (Union[ProxyTag,ProxyTags,Sequence[ProxyTag]]): Member proxy tags.
            keep_proxy (bool): Whether to keep the proxy prefix and/or suffix for this member's
                proxied messages.
            visibility (Privacy): Whether this member is visible to others (i.e. in member lists).
                Default is public.
            name_privacy (Privacy): Whether the member name is visible to others or only the display
                name. Default is public.
            description_privacy (Privacy): Whether the member description is visible to others.
                Default is public.
            birthday_privacy (Privacy): Whether the member birthday is visible to others. Default
                is public.
            pronoun_privacy (Privacy): Whether the member pronouns are visible to others. Default
                is public.
            avatar_privacy (Privacy): Whether the member avatar is visible to others. Default is
                public.
            metadata_privacy (Privacy): Whether the member's metadata (i.e. created timestamp,
                message count) is visible to others. Default is public.
        """
        name = _PATCHABLE_MEMBER_KEYS["name"](name)
        self._check_update_keys("new_member()", kwargs, _PATCHABLE_MEMBER_KEYS,
            require_at_least_one_arg=False)
        payload = {"name": name}
        privacies = {}
        for key, value in kwargs.items():
            json_value = _PATCHABLE_MEMBER_KEYS[key](value)
            if key in _PRIVACY_ASSOCIATED_KEYS:
                privacies[key] = json_value
            else:
                payload[key] = json_value
        if privacies: payload["privacy"] = privacies
        return self._request_something(
            "POST",
            "{SERVER}/members",
            Member,
            200,
            GENERIC_ERROR_CODE_LOOKUP,
            payload=payload,
        )

    @_async_mode_handler
    def new_group(self, name: str, **kwargs) -> Group:
        """Create a new group

        Args:
            name: Name of the group. 100 character limit.

        Keyword args:
            display_name (str): Group display name (100 character limit).
            description (str): Group description (1,000 character limit).
            icon (str): Group icon url (256 character limit).
            banner (str): Group banner url (256 character limit).
            color (Color): Group color.
            name_privacy (Privacy): Whether the group name is visible to others or only the display
                name. Default is public.
            description_privacy (Privacy): Whether the group description is visilbe to others.
                Default is public.
            icon_privacy (Privacy): Whether the group icon is visible to others. Default is public.
            list_privacy (Privacy): Whether the group member list is visible to others. Default is
                public.
            metadata_privacy (Privacy): Whether the groups's metadata (i.e. created timestamp) is
                visible to others. Default is public.
            visibility (Privacy): Whether this group is visible to others (i.e. in group lists).
                Default is public.
        """
        name = _PATCHABLE_GROUP_KEYS["name"](name)
        self._check_update_keys("new_group()", kwargs, _PATCHABLE_GROUP_KEYS,
            require_at_least_one_arg=False)
        payload = {"name": name}
        privacies = {}
        for key, value in kwargs.items():
            json_value = _PATCHABLE_GROUP_KEYS[key](value)
            if key in _PRIVACY_ASSOCIATED_KEYS:
                privacies[key] = json_value
            else:
                payload[key] = json_value
        if privacies: payload["privacy"] = privacies
        return self._request_something(
            "POST",
            "{SERVER}/groups",
            Group,
            200,
            GENERIC_ERROR_CODE_LOOKUP,
            payload=payload,
        )

    @_async_mode_handler
    def new_switch(self, members, timestamp=None) -> Switch:
        """Log a new switch.

        Args:
            members (Sequence[Union[MemberId,Member]]): List of members in this switch.

        Keyword args:
            timestamp (Timestamp): Time of switch. ``None`` for "now" (default).
        """
        payload = {}
        payload["members"] = _PATCHABLE_SWITCH_KEYS["members"](members)
        if timestamp is not None:
            payload["timestamp"] = _PATCHABLE_SWITCH_KEYS["timestamp"](timestamp)
        return self._request_something(
            "POST",
            "{SERVER}/systems/{system_ref}/switches",
            Switch,
            200,
            GENERIC_ERROR_CODE_LOOKUP,
            payload=payload,
        )




from http.client import responses as RESPONSE_CODES

class PluralKitException(Exception):
    """Base exception class for pluralkit.py
    """

class HTTPError(PluralKitException):
    """General PluralKit exception class for unexpected HTTP codes (e.g. 500).
    """

class GenericBadRequest(PluralKitException):
    """Thrown when the API server informs the client that the request was malformed (400).
    """

# Exceptions related to "not found"-related errors

class NotFound(PluralKitException):
    """Thrown when the resource cannot be found (404).
    """

class MemberNotFound(NotFound):
    """Thrown when the Member ID is apparently not in PluralKit's database (404).
    """

class SystemNotFound(NotFound):
    """Thrown when the System ID is apparently not in PluralKit's database (404).
    """

class SwitchNotFound(NotFound):
    """Thrown when the Switch ID is apparently not in PluralKit's database (404).
    """

class MessageNotFound(NotFound):
    """Thrown when the message ID is apparently not in PluralKit's database (404).
    """

class GroupNotFound(NotFound):
    """Thrown when the Group ID is apparently not in PluralKit's database (404).
    """

class GuildNotFound(NotFound):
    """Thrown when the member or system has no guild settings for a given guild (404).
    """

# Exceptions related to lack of authorization

class Unauthorized(PluralKitException):
    """Thrown when the authorization token passed to PluralKit's API is invalid or missing (403).
    """

class NotOwnSystem(Unauthorized):
    """Thrown when the client doesn't have access to a system's private info (403).
    """

class NotOwnMember(Unauthorized):
    """Thrown when the client doesn't have access to a member's private info (403).
    """

class NotOwnGroup(Unauthorized):
    """Thrown when the client doesn't have access to a group's private info (403).
    """

# Python <= 3.8 RTD fix
def merge(d1, d2): return {**d1, **d2}

GENERIC_ERROR_CODE_LOOKUP = {
    400: GenericBadRequest,
    403: Unauthorized,
    404: NotFound,
}

SYSTEM_ERROR_CODE_LOOKUP = merge(GENERIC_ERROR_CODE_LOOKUP, {
    401: Unauthorized,
    403: NotOwnSystem,
    404: SystemNotFound,
})

MEMBER_ERROR_CODE_LOOKUP = merge(GENERIC_ERROR_CODE_LOOKUP, {
    401: Unauthorized,
    403: NotOwnMember,
    404: MemberNotFound,
})

GROUP_ERROR_CODE_LOOKUP = merge(GENERIC_ERROR_CODE_LOOKUP, {
    401: Unauthorized,
    403: NotOwnGroup,
    404: GroupNotFound,
})

MESSAGE_ERROR_CODE_LOOKUP = merge(GENERIC_ERROR_CODE_LOOKUP, {
    401: Unauthorized,
    404: MessageNotFound,
})

SWITCH_ERROR_CODE_LOOKUP = merge(GENERIC_ERROR_CODE_LOOKUP, {
    401: Unauthorized,
    403: NotOwnSystem,
    404: SwitchNotFound,
})

GUILD_ERROR_CODE_LOOKUP = merge(GENERIC_ERROR_CODE_LOOKUP, {
    401: Unauthorized,
    404: GuildNotFound,
})




from .client import Client
from .models import (
    SystemId, System, SystemSettings, SystemGuildSettings,
    MemberId, Member, MemberGuildSettings,
    AutoproxySettings,
    GroupId, Group,
    SwitchId, Switch,
    Message,
    Birthday,
    Color,
    ProxyTag, ProxyTags,
    Timestamp,
    Timezone,
    Privacy, AutoproxyMode,
)

from .errors import (
    PluralKitException,
    HTTPError,
    GenericBadRequest,
    NotFound,
        SystemNotFound,
        MemberNotFound,
        GroupNotFound,
        SwitchNotFound,
        MessageNotFound,
        GuildNotFound,
    Unauthorized,
        NotOwnSystem,
        NotOwnMember,
        NotOwnGroup,
)




from collections import namedtuple

VersionInfo = namedtuple("VersionInfo", "major minor build")
version_info = VersionInfo(
    major=1,
    minor=2,
    build=0,
)

__version__ = f"{version_info.major}.{version_info.minor}.{version_info.build}"




from __future__ import annotations

from dataclasses import dataclass
from string import ascii_lowercase as ALPHABET
from enum import Enum
from datetime import datetime, timedelta, tzinfo
import re
from typing import (
    Any,
    Union, Optional,
    Tuple, List, Set, Sequence, Dict,
    Generator,
)
import warnings

import pytz
import colour

from .errors import *

# Enums

class Privacy(Enum):
    """Represents the privacies accepted by PluralKit.

    In general, default privacies are public.
    """
    PUBLIC = "public"
    PRIVATE = "private"
    #UNKNOWN = None # legacy, effectively resets privacy to "public"

    def json(self): return self.value

class AutoproxyMode(Enum):
    """Represents the autproxy modes.
    """
    OFF = "off"
    FRONT = "front"
    LATCH = "latch"
    MEMBER = "member"

    def json(self): return self.value

def _to_json(value):
    """Robust method to deep convert Model objects
    """
    if hasattr(value, "json"):
        return value.json()
    
    if isinstance(value, (list, set, tuple)):
        return [_to_json(v) for v in value]
    
    if isinstance(value, (dict,)):
        return {k: _to_json(v) for k, v in value.items()}
    
    return value

# Base class for all models

class Model:
    """Base class for all models.
    """

    def json(self):
        """Return a JSON object representing this model.
        """
        model = {}
        for k, v in self.__dict__.items():
            if k.startswith("_"): continue
            if v is None: continue
            model[k] = _to_json(v)

        return model

    def __init__(self, json, ignore_keys=None):
        """Simple way to convert from API JSON object to the superclass
        """
        if ignore_keys is None: ignore_keys = ()
        cls = self.__class__

        for key, value in json.items():
            if key in ignore_keys: continue
            if key not in cls.__annotations__ and key not in _KEY_TRANSFORMATIONS:
                msg = f"unexpected key {key!r} in JSON object for {cls.__name__!r} construction"
                warnings.warn(msg)

            # convert PluralKit API key names to pluralkit.py attribute names
            # and convert to proper Models if necessary
            if key in _VALUE_TRANSFORMATIONS:
                Constructor = _VALUE_TRANSFORMATIONS[key]
                value = Constructor(value) if value is not None else None
            key = _KEY_TRANSFORMATIONS.get(key, key)

            self.__dict__[key] = value

# IDs

class PluralKitId(Model):
    """Base class for PluralKit IDs.
    """
    uuid: Optional[str]
    id: Optional[str]

    _six_character_pattern = re.compile(
        "^([a-z]{{3}}-?[a-z]{{3}})$",
        re.IGNORECASE,
    )

    __slots__ = ["uuid", "id"]

    def _check_id(self, id: str) -> bool:
        # 5-character IDs
        if len(id) == 5 and all(c in ALPHABET for c in id): return True
        # 6-character IDs
        return bool(self._six_character_pattern.match(id))

    def _check_uuid(self, uuid):
        pattern = r"[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}"
        result = re.match(pattern, uuid)
        return bool(result)

    def __init__(self, id=None, uuid=None):
        if uuid is None and id is None:
            raise ValueError(f"{self.CONTEXT} ID object must include at least one of: id, uuid")

        # accept any order/combination of inputs
        object.__setattr__(self, "id", None)
        object.__setattr__(self, "uuid", None)
        if id is not None:
            if self._check_id(id):
                if len(id) >= 6:
                    id = id.lower()
                    id = f"{id[:3]}-{id[-3:]}"
                object.__setattr__(self, "id", id)
            elif self._check_uuid(id):
                object.__setattr__(self, "uuid", id)
            else:
                raise ValueError(f"Malformed id given: {id!r}")
        if uuid is not None:
            if self._check_uuid(uuid):
                object.__setattr__(self, "uuid", uuid)
            elif self._check_id(uuid):
                if len(uuid) >= 6:
                    uuid = uuid.lower()
                    uuid = f"{uuid[:3]}-{uuid[-3:]}"
                object.__setattr__(self, "id", uuid)
            else:
                raise ValueError(f"Malformed uuid given: {uuid!r}")

    def __setattr__(self, name, value):
        msg = f"cannot assign to field {name!r}"
        raise AttributeError(msg)

    def __str__(self):
        return f"{self.uuid}" if self.uuid is not None else f"{self.id}"

    def __repr__(self):
        attrs = f"{self.id!r}"
        if self.uuid is not None: attrs += f", {self.uuid!r}"
        return f"{self.__class__.__name__}({attrs})"

    def json(self) -> str:
        return (
            f"{self.uuid}"
            if self.uuid is not None
            else f"{self.id.replace('-', '')}"
        )

class MemberId(PluralKitId):
    """Member IDs.
    """
    CONTEXT = "Member"

class SystemId(PluralKitId):
    """System IDs.
    """
    CONTEXT = "System"

class GroupId(PluralKitId):
    """Group IDs.
    """
    CONTEXT = "Group"

class SwitchId(PluralKitId):
    """Switch IDs.

    Switches don't have five-letter IDs, so this must be given the full switch UUID.
    """
    uuid: str
    CONTEXT = "Switch"
    __slots__ = ["uuid"]

    def __init__(self, uuid):
        if uuid is None:
            raise ValueError(f"{self.CONTEXT} ID object must include uuid")

        object.__setattr__(self, "uuid", uuid)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.uuid!r})"

# Primitives

class Color(colour.Color, Model):
    """Represents a color.

    This class is initialized in the same way that a `colour.Color`_ object is. It may also take a
    `colour.Color`_ object directly.
    
    .. _`colour.Color`: https://pypi.org/project/colour/#instantiation
    """
    def __init__(self, *args, **kwargs):
        if len(args) == 1 and len(kwargs) == 0:
            arg = args[0]
            if isinstance(arg, colour.Color):
                colour.Color.__init__(self, args[0].hex_l)
            elif isinstance(arg, str):
                # add octothorp to string if not there
                pattern = r"^([A-Fa-f0-9]{6})$"
                if re.search(pattern, arg):
                    arg = "#" + arg
                colour.Color.__init__(self, arg)
            return

        colour.Color.__init__(self, *args, *kwargs)

    def __str__(self):
        return self.hex_l[1:]

    json = __str__

class Timestamp(Model):
    """Represents a PluralKit UTC timestamp.

    This class works by wrapping around a `datetime` object. Use ``ts.datetime`` to access it, for
    any `Timestamp` ``ts``.

    This class may be initialized in the same way that a `datetime` object is. It may also take a
    `datetime` object, a `Timestamp` object, or an ISO 8601 formatted string directly.
    """
    def __init__(self, dt: Optional[datetime]=None, *,
        year: Optional[int]=None,
        month: Optional[int]=None,
        day: Optional[int]=None,
        hour: int=0,
        minute: int=0,
        second: int=0,
        microsecond: int=0
    ):
        if dt is None and any(arg is None for arg in (year, month, day)):
            msg = (
                f"{self.__class__.__name__} is missing required arguments. Either provide a "
                f"datetime.datetime or ISO 8601 formatted string via the first positional "
                f"argument, or provide the year, month, and day through the respective keyword "
                f"arguments."
            )
            raise TypeError(msg)

        if dt is not None:
            if isinstance(dt, datetime):
                if dt.tzinfo is not None:
                    self.datetime = dt.astimezone(pytz.utc)
                else:
                    self.datetime = dt.replace(tzinfo=pytz.utc)
            elif isinstance(dt, str):
                try:
                    self.datetime = datetime.strptime(dt, r"%Y-%m-%dT%H:%M:%S.%fZ")
                except ValueError:
                    self.datetime = datetime.strptime(dt, r"%Y-%m-%dT%H:%M:%SZ")
            elif isinstance(dt, Timestamp):
                self.datetime = dt.datetime
            else:
                msg = (
                    f"{self.__class__.__name__} takes either a datetime.datetime object or "
                    f"ISO 8601 formatted string as the first positional argument. Given "
                    f"type(dt)={type(dt)!r}"
                )
                raise TypeError(msg)

        else:
            self.datetime = datetime(year, month, day, hour, minute, second, microsecond)
            self.datetime = self.datetime.replace(tzinfo=pytz.utc)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.json()!r})"

    def __str__(self):
        return (
            f"{self.year:04d}-{self.month:02d}-{self.day:02d} "
            f"{self.hour:02d}:{self.minute:02d}:{self.second:02d} UTC"
        )
    
    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.json() == other.json()
        elif isinstance(other, datetime):
            if other.tzinfo is None:
                return self.datetime == other.replace(tzinfo=pytz.utc) # assume UTC
            else:
                return self.datetime == other

        return NotImplemented
    
    def __ne__(self, other):
        x = self.__eq__(other)

        if x is NotImplemented:
            return NotImplemented
        else:
            return not x

    def __lt__(self, other):
        if isinstance(other, self.__class__):
            return self.datetime < other.datetime
        elif isinstance(other, datetime):
            if other.tzinfo is None:
                return self.datetime < other.replace(tzinfo=pytz.utc) # assume UTC
            else:
                return self.datetime < other

        return NotImplemented

    def __le__(self, other):
        if isinstance(other, self.__class__):
            return self.datetime <= other.datetime
        elif isinstance(other, datetime):
            if other.tzinfo is None:
                return self.datetime <= other.replace(tzinfo=pytz.utc) # assume UTC
            else:
                return self.datetime <= other

        return NotImplemented

    def __gt__(self, other):
        if isinstance(other, self.__class__):
            return self.datetime > other.datetime
        elif isinstance(other, datetime):
            if other.tzinfo is None:
                return self.datetime > other.replace(tzinfo=pytz.utc) # assume UTC
            else:
                return self.datetime > other

        return NotImplemented

    def __ge__(self, other):
        if isinstance(other, self.__class__):
            return self.datetime >= other.datetime
        elif isinstance(other, datetime):
            if other.tzinfo is None:
                return self.datetime >= other.replace(tzinfo=pytz.utc) # assume UTC
            else:
                return self.datetime >= other

        return NotImplemented

    @property
    def year(self):
        return self.datetime.year

    @year.setter
    def year(self, value):
        self.datetime = self.datetime.replace(year=value)

    @property
    def month(self):
        return self.datetime.month

    @month.setter
    def month(self, value):
        self.datetime = self.datetime.replace(month=value)

    @property
    def day(self):
        return self.datetime.day

    @day.setter
    def day(self, value):
        self.datetime = self.datetime.replace(day=value)

    @property
    def hour(self):
        return self.datetime.hour

    @hour.setter
    def hour(self, value):
        self.datetime = self.datetime.replace(hour=value)

    @property
    def minute(self):
        return self.datetime.minute

    @minute.setter
    def minute(self, value):
        self.datetime = self.datetime.replace(minute=value)

    @property
    def second(self):
        return self.datetime.second

    @second.setter
    def second(self, value):
        self.datetime = self.datetime.replace(second=value)

    @property
    def microsecond(self):
        return self.datetime.microsecond

    @microsecond.setter
    def microsecond(self, value):
        self.datetime = self.datetime.replace(microsecond=value)

    def json(self) -> str:
        """Convert this timestamp to the ISO 8601 format that PluralKit uses internally.
        """
        return (
            f"{self.year:04d}-{self.month:02d}-{self.day:02d}"
            f"T{self.hour:02d}:{self.minute:02d}:{self.second:02d}.{self.microsecond:06d}Z"
        )

class Birthday(Timestamp):
    """Represents a birthday.

    This model inherits from `Timestamp`.
    """
    def __str__(self):
        if self.hidden_year:
            return self.datetime.strftime("%b %d")
        else:
            return self.datetime.strftime("%b %d, ") + f"{self.year:04d}"

    @property
    def hidden_year(self) -> bool:
        """Whether this birthday's year is hidden.

        If set to ``False``, sets the birthday's year to ``0001``, `which internally represents a
        hidden year in PluralKit's API`_.

        .. _`which internally represents a hidden year in PluralKit's API`:
            https://pluralkit.me/api/#member-model
        """
        return self.year in (1, 4)

    @hidden_year.setter
    def hidden_year(self, value: bool):
        if value == True:
            self.year = 1
        else:
            pass # nothing one can do ?

    @staticmethod
    def from_json(bd: str):
        """Takes in a string (as returned by the API) and returns the `Birthday`.

        Args:
            bd: The ``YYYY-MM-DD`` formatted string representing the birthdate.

        Returns:
            Birthday: The corresponding birthday.
        """
        return Birthday(datetime.strptime(bd, r"%Y-%m-%d"))

    def json(self) -> str:
        """Returns the ``YYYY-MM-DD`` formatted birthdate.
        """
        return f"{self.year:04d}-{self.month:02d}-{self.day:02d}"

class Timezone(Model):
    """Represents a tzdb time zone.

    This class is initialized in the same way that `pytz.timezone`_ initializes `tzinfo` objects.
    It may also take a `tzinfo` object directly.

    Hint:
        `Here is a link to a list of tz database time zones`_

    Args:
        tz (Union[str,tzinfo]): The timezone, either as a string or as a `tzinfo` (e.g. from
            `pytz.timezone`_).

    .. _`pytz.timezone`: http://pytz.sourceforge.net/
    .. _`Here is a link to a list of tz database time zones`:
        https://en.wikipedia.org/wiki/List_of_tz_database_time_zones
    """
    def __init__(self, tz):
        if isinstance(tz, tzinfo):
            self.tz = tz
        else:
            self.tz = pytz.timezone(tz)

    def __eq__(self, other):
        return self.tz.zone == other.tz.zone

    def __repr__(self):
        return f"{self.__class__.__name__}({self.tz.zone!r})"

    def json(self):
        """Returns the string representation of this timezone as expected by the API.
        """
        return self.tz.zone

# Settings

class MemberGuildSettings(Model):
    """Member settings for a specific server.

    Attributes:
        display_name: The member's display name in the guild.
        avatar_url: The URL of the member's avatar image in the guild.
    """
    display_name: Optional[str]
    avatar_url: Optional[str]

class SystemGuildSettings(Model):
    """System settings for a specific server.

    Attributes:
        proxying_enabled: Whether proxying is enabled in the given server.
        tag: The system's tag (appended to the server username) for the given server.
        tag_enabled: Whether or not the system tag is shown in this server.
    """
    proxying_enabled: bool
    tag_enabled: bool
    tag: Optional[str]

class SystemSettings(Model):
    """Represents a system's settings.

    Attributes:
        timezone: The system's timezone.
        pings_enabled: Whether this system has pings enabled.
        latch_timeout: System's autoproxy latch timeout.
        member_default_private: Whether members created through the bot have privacy settings set
            to private by default.
        group_default_private: Whether groups created through the bot have privacy settings set to
            private by default.
        show_private_info: Whether the bot shows the system's own private information without
            requiring a ``-private`` flag.
        member_limit: System member limit, usually 1000.
        group_limit: System group limit, usually 250.
    """
    timezone: Timezone
    pings_enabled: bool
    latch_timeout: Optional[int]
    member_default_private: bool
    group_default_private: bool
    show_private_info: bool
    member_limit: int
    group_limit: int

    def __init__(self, json):
        Model.__init__(self, json, ("description_templates"))

class AutoproxySettings(Model):
    """Represents a system's autoproxy settings.

    Attributes:
        autoproxy_mode: The system's autoproxy mode.
        autoproxy_member: ID of current autoproxy member. (None if autoproxy mode is set to
            ``front``.)
        last_latch_timestamp: Timestamp of last message. (None if autoproxy mode isn't set to
            ``latch``.)
    """
    autoproxy_mode: AutoproxyMode
    autoproxy_member: Optional[MemberId]
    last_latch_timestamp: Optional[Timestamp]

# Proxy tags

class ProxyTag(Model):
    """Represents a single PluralKit proxy tag.

    Hint:
        A ProxyTag object can be called to see if it would match a message: ::

            >>> pt = ProxyTag("{", "}")
            >>> pt("This is an example.")
            False
            >>> pt("{This is another example.}")
            True

    Args:
        prefix: Prefix that will enclose proxied messages.
        suffix: Suffix that will enclose proxied messages.

    Keyword args:
        proxy_tag: Dictionary representing a proxy tag. Must have at least one of ``prefix`` or
            ``suffix`` as keys. The ``prefix`` and ``suffix`` args will overrule this dict.

    Important:
        At least one of the ``suffix`` or ``prefix`` arguments must be passed.
    
    Attributes:
        prefix (Optional[str]): Prefix that will enclose proxied messages.
        suffix (Optional[str]): Suffix that will enclose proxied messages.
    """
    def __init__(self,
        prefix: Optional[str]=None,
        suffix: Optional[str]=None,
        *,
        proxy_tag: Optional[Dict[str,str]]=None,
    ):

        if proxy_tag is not None:
            prefix = prefix or proxy_tag["prefix"]
            suffix = suffix or proxy_tag["suffix"]

        assert prefix or suffix, \
            "A valid proxy tag must have at least one of the prefix or suffix defined."
        self.prefix = prefix
        self.suffix = suffix
    
    def __eq__(self, other):
        return self.prefix == other.prefix and self.suffix == other.suffix
    
    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        prefix = "" if not self.prefix else f"prefix={repr(self.prefix)}"
        suffix = "" if not self.suffix else f"suffix={repr(self.suffix)}"
        attrs = ", ".join(a for a in (prefix, suffix) if a)
        return (
            f"{self.__class__.__name__}({attrs})"
        )

    def __call__(self, message: str) -> bool:
        """Determine if a given message would be proxied under this proxy tag.
        
        Args:
            message: Message to parse.
        """
        message = message.strip()
        return (True if not self.prefix else message.startswith(self.prefix)) \
            and (True if not self.suffix else message.endswith(self.suffix))

    def json(self) -> Dict[str,Optional[str]]:
        """Return the JSON object representing this proxy tag as a Python `dict`.
        """
        return {
            "prefix": self.prefix,
            "suffix": self.suffix,
        }

class ProxyTags(Model):
    """Represents a set of PluralKit proxy tags.

    Hint:
        A ProxyTags object can be called to see if it would match a message: ::

            >>> pt_1 = ProxyTag("{", "}")
            >>> pt_2 = ProxyTag("A:")
            >>> pts = ProxyTags([pt_1, pt_2])
            >>> pts("{This is an example.}")
            True
            >>> pts("A: This is another example.")
            True

    Hint:
        ProxyTags objects can be iterated or indexed to yield its underlying `ProxyTag` objects.    
    
    Args:
        proxy_tags: A sequence of `ProxyTag` objects.
    """
    def __init__(self, proxy_tags: Optional[Generator[ProxyTag,None,None]]=None):
        self._proxy_tags: Tuple[ProxyTag,...]
        if proxy_tags is None:
            self._proxy_tags = tuple()
        else:
            self._proxy_tags = tuple(proxy_tags)

    def __repr__(self):
        return f"{self.__class__.__name__}({list(self._proxy_tags)!r})"

    def __iter__(self):
        for proxy_tag in self._proxy_tags:
            yield proxy_tag

    def __getitem__(self, index):
        return self._proxy_tags[index]
    
    def __eq__(self, other):
        return set(self._proxy_tags) == set(other._proxy_tags)
    
    def __ne__(self, other):
        return not self.__eq__(other)

    def __call__(self, message: str) -> bool:
        """Determine if a given message would be proxied under this set of proxy tags.
        
        Args:
            message: Message to parse.
        """
        return any(proxy_tag.match(message) for proxy_tag in self)

    def json(self) -> List[Dict[str,str]]:
        """Return the JSON object representing this proxy tag as a list of Python `dict`.
        """
        return [proxy_tag.json() for proxy_tag in self]

# Member, System, Group, Switch, and Message

class Member(Model):
    """Represents a PluralKit system member.

    Attributes:
        id: The member reference ID.
        system: The ID of the system this member belongs to.
        name: The member's name.
        created: The member's creation timestamp.
        display_name: The member's display name.
        description: The member's description.
        color: The member's color.
        birthday: The member's birthday.
        pronouns: The member's pronouns.
        avatar_url: The member's main avatar url that appears uncropped on the member card.
        webhook_avatar_url: The member's avatar url used for proxied messages instead of the main
            avatar.
        banner: The member's banner url.
        proxy_tags: The member's proxy tags.
        keep_proxy (bool): Whether the member's proxy tags remain in the proxied message or not.
        name_privacy: Whether the member name is visible to others or only the display name.
        description_privacy: Whether this member's description is visible to others.
        birthday_privacy: Whether the member's birthday is visible to others.
        pronoun_privacy: Whether the member's pronouns are visible to others.
        avatar_privacy: Whether the member's avatar is visible to others.
        metadata_privacy: Whether the member's metadata (i.e. creation timestamp, message count) is
            visible to others.
        visibility: Whether this member is visible to others (i.e. in member lists).
        autoproxy_enabled: Whether this member has autoproxy enabled. `None` if the member is not
            from the client's system.
        message_count: Member message count. `None` if the member's metadata privacy is set to
            private and the member is not from the client's system.
        last_message_timestamp: Timestamp of member's last message. `None` if the member's metadata
            privacy is set to private and the member is not from the client's system.
        tts: Whether this member has enabled tts or not
    """
    id: MemberId
    name: str
    created: Timestamp
    name_privacy: Optional[Privacy]
    description_privacy: Optional[Privacy]
    birthday_privacy: Optional[Privacy]
    pronoun_privacy: Optional[Privacy]
    avatar_privacy: Optional[Privacy]
    metadata_privacy: Optional[Privacy]
    visibility: Optional[Privacy]
    display_name: Optional[str]
    description: Optional[str]
    color: Optional[Color]
    birthday: Optional[Birthday]
    pronouns: Optional[str]
    avatar_url: Optional[str]
    webhook_avatar_url: Optional[str]
    keep_proxy: bool
    proxy_tags: Optional[ProxyTags]
    system: SystemId
    banner: Optional[str]
    autoproxy_enabled: Optional[bool]
    message_count: Optional[int]
    last_message_timestamp: Optional[Timestamp]
    tts: Optional[bool]

    def __str__(self):
        return f"{self.id!s}"
    
    def __eq__(self, other):
        return self.id == other.id
    
    def __ne__(self, other):
        return not self.__eq__(other)

    def __init__(self, json):
        ignore_keys = ("uuid", "id", "privacy",)
        Model.__init__(self, json, ignore_keys)
        # fix up the remaining keys
        self.id = MemberId(id=json["id"], uuid=json["uuid"]) 
        # incorporate privacy keys when given
        if json.get("privacy") is not None:
            for key, value in json["privacy"].items():
                self.__dict__[key] = Privacy(value) if value is not None else None
        else:
            for key, _ in self.__class__.__annotations__.items():
                if "privacy" in key:
                    self.__dict__[key] = None # unknown         

class System(Model):
    """Represents a PluralKit system.

    Attributes:
        id: The system reference ID.
        name: The system's name.
        description: The system's description.
        created: The system's creation timestamp.
        tag: The system's tag appended to display names.
        pronouns: The system's pronouns.
        avatar_url: The system's avatar url.
        banner: The system's banner url.
        tz: The system's tzdb timezone.
        color: The system's color.
        description_privacy: Whether the system's description is visible to others.
        pronoun_privacy: Whether the system's pronouns are visible to others.
        member_list_privacy: Whether the system's member list is visible to others.
        group_list_privacy: Whether the system's group list is visible to others.
        front_privacy: Whether the system's current fronter information is visible to others.
        front_history_privacy: Whether the system's front history is visible to others.
    """
    id: SystemId
    created: Timestamp
    name: Optional[str]
    description: Optional[str]
    tag: Optional[str]
    avatar_url: Optional[str]
    tz: Timezone
    description_privacy: Optional[Privacy]
    pronoun_privacy: Optional[Privacy]
    member_list_privacy: Optional[Privacy]
    group_list_privacy: Optional[Privacy]
    front_privacy: Optional[Privacy]
    front_history_privacy: Optional[Privacy]
    pronouns: Optional[str]
    banner: Optional[str]
    color: Optional[Color]

    def __str__(self):
        return f"{self.id!s}"
    
    def __eq__(self, other):
        return self.id == other.id
    
    def __ne__(self, other):
        return not self.__eq__(other)

    def __init__(self, json):
        ignore_keys = ("privacy", "webhook_url", "id", "uuid",)
        Model.__init__(self, json, ignore_keys)
        # fix up the remaining keys
        self.id = SystemId(id=json["id"], uuid=json["uuid"])
        # incorporate privacy keys when given
        if json.get("privacy") is not None:
            for key, value in json["privacy"].items():
                self.__dict__[key] = Privacy(value) if value is not None else None
        else:
            for key, _ in self.__class__.__annotations__.items():
                if "privacy" in key:
                    self.__dict__[key] = None # unknown

class Group(Model):
    """Represents a PluralKit system group.

    Attributes:
        id: The group reference ID.
        system: The ID of the system this group belongs to.
        name: Group name.
        display_name: Group display name.
        description: Group description.
        icon: Group icon url.
        banner: Group banner url.
        color: Group color.
        created: The group's creation timestamp.
        name_privacy: Whether the group name is visible to others or only the display name.
        description_privacy: Whether the group description is visilbe to others.
        icon_privacy: Whether the group icon is visible to others.
        list_privacy: Whether the group member list is visible to others.
        metadata_privacy: Whether the groups's metadata (i.e. created timestamp) is visible to
            others.
        visibility: Whether this group is visible to others (i.e. in group lists).
    """
    id: Optional[GroupId]
    system: SystemId
    name: str
    display_name: Optional[str]
    description: Optional[str]
    icon: Optional[str]
    banner: Optional[str]
    color: Optional[Color]
    name_privacy: Privacy
    description_privacy: Privacy
    icon_privacy: Privacy
    list_privacy: Privacy
    metadata_privacy: Privacy
    visibility: Privacy
    created: Timestamp

    def __str__(self):
        return f"{self.id!s}"
    
    def __eq__(self, other):
        return self.id == other.id

    def __init__(self, json):
        ignore_keys = ("uuid", "id", "privacy",)
        Model.__init__(self, json, ignore_keys)
        # fix up the remaining keys
        self.id = GroupId(id=json["id"], uuid=json["uuid"])
        for key, value in json["privacy"].items():
            self.__dict__[key] = Privacy(value)

class Switch(Model):
    """Represents a switch event.

    Note:
        ``members`` can either be a list of `Member` models or a list of `MemberId` objects,
        depending on the client method used.

        In particular, switch models from `Client.get_switches` carry `MemberId` objects, whereas
        switch models from  `~Client.get_switch`, `~Client.new_switch`, and `~Client.update_switch`
        carry full `Member` objects.

    Attributes:
        id: Switch's unique universal identifier (uuid).
        timestamp: Timestamp of the switch.
        members: Members involved.
    """
    id: SwitchId
    timestamp: Timestamp
    members: Union[Sequence[Member],Sequence[MemberId]]

    def __init__(self, json):
        ignore_keys = ("id", "members",) # "members" key is tricky
        Model.__init__(self, json, ignore_keys)
        # ...
        self.id = SwitchId(json["id"])
        if all(isinstance(m, str) for m in json["members"]):
            self.members = [MemberId(m) for m in json["members"]]
        else:
            self.members = [Member(m) for m in json["members"]]
    
    def __eq__(self, other):
        return self.__dict__ == other.__dict__
    
    def __ne__(self, other):
        return not self.__eq__(other)

class Message(Model):
    """Represents a proxied message.

    Attributes:
        timestamp: Timestamp of the message.
        id: The ID of the Discord message sent by the webhook.
        original: The ID of the (presumably deleted) original Discord message sent by the account.
        sender: The user ID of the account that sent the message.
        channel: The ID of the channel the message was sent to.
        guild: The ID of the guild the message was sent in.
        system: The system that proxied the message. ``None`` if system was deleted.
        member: The member that proxied the message. ``None`` if member was deleted.
    """
    timestamp: Timestamp
    id: int
    original: int
    sender: int
    channel: int
    guild: int
    system: Optional[System]
    member: Optional[Member]
    
    def __eq__(self, other):
        return self.id == other.id
    
    def __ne__(self, other):
        return not self.__eq__(other)

    def __init__(self, json):
        ignore_keys = ("id", "system", "member",)
        Model.__init__(self, json, ignore_keys)
        # fix the remaining keys
        self.system = None if json["system"] is None else System(json["system"])
        self.member = None if json["member"] is None else Member(json["member"])
        self.id = int(json["id"])

def _proxy_tags_processor(proxy_tags):
    if not proxy_tags: return proxy_tags
    return ProxyTags([ProxyTag(proxy_tag=pt) for pt in proxy_tags])

# the following maps direct how to change the JSON objects as given by the API
# to make them ready for use (e.g. Python-friendly)
# see Model.__init__ for how this is used

# [name given by API] -> [new pk.py (Python-friendly) name]
_KEY_TRANSFORMATIONS = {
    # (So far, all key names in the API are Python-friendly)
}

# [name given by API] -> [constructor to use on this object]
_VALUE_TRANSFORMATIONS = {
    "system": SystemId, # can also be full system object (for Message)
    "color": lambda c: None if c is None else Color(c),
    "proxy_tags": _proxy_tags_processor,
    "created": Timestamp,
    "timestamp": Timestamp,
    "last_message_timestamp": Timestamp,
    "message_count": int,
    "channel": int,
    "original": int,
    "sender": int,
    "guild": int,
    "timezone": Timezone,
}

# Patchable keys, along with their respective checks
# These methods are for the client's patch (update) methods

def _max_string_length(context, max_len, null_allowed=True):
    def check(value):
        if null_allowed and value is None: return None
        if isinstance(value, str) and len(value) <= max_len: return value
        msg = (
            f"value for {context!r} must be shorter than {max_len} characters "
            f"(counted {len(value)})"
        )
        raise ValueError(msg)
    return check

def _check_color(c):
    if isinstance(c, Color): return c.json()
    c = Color(c) # will throw error if malformed
    return c.json()

def _check_privacy(p):
    if p is None: return None # null allowed
    if isinstance(p, Privacy): return p.value
    p = Privacy(p) # will throw error if malformed
    return p.value

def _check_timestamp(t):
    if isinstance(t, Timestamp): return t.json()
    t = Timestamp(t) # will throw error if malformed
    return t.json()

def _check_birthday(b):
    if b is None: return None # null allowed
    if isinstance(b, Birthday): return b.json()
    if isinstance(b, Timestamp): return Birthday(b).json()
    b = Birthday(b) # will throw error if malformed
    return b.json()

def _check_timezone(tz):
    if tz is None: return "UTC"
    if isinstance(tz, Timezone): return tz.json()
    tz = Timezone(tz) # will throw error if malformed
    return tz.json()

def _check_proxy_tags(pts):
    """Allowed: ProxyTag, ProxyTags, Sequence[ProxyTag]
    """
    if isinstance(pts, ProxyTags): return pts.json()
    if isinstance(pts, ProxyTag): return [pts.json()],
    try:
        json = []
        for pt in pts:
            if not isinstance(pt, ProxyTag):
                pt = ProxyTag(proxy_tag=pt)
            json.append(pt.json())
        return json
    except (TypeError, ValueError):
        msg = (
            f"Could not cast {pts!r} to ProxyTags. "
            f"Please pass in a ProxyTags object, a ProxyTag object, or an "
            f"iterable of ProxyTag objects."
        )
        raise ValueError(msg)

def _check_members(members):
    """For `Client.update_switch`
    """
    # special case for one member
    if isinstance(members, MemberId): return [str(members)]
    if isinstance(members, Member): return [str(members.id)]
    # otherwise
    try:
        json = []
        for m in members:
            if isinstance(m, Member):
                json.append(str(m.id))
            elif isinstance(m, MemberId):
                json.append(str(m))
            else:
                m = MemberId(m)
                json.append(str(m))
        return json
    except (TypeError, ValueError):
        msg = (
            f"Could not cast {members!r} to a list of MemberId or Member. "
            f"Please pass in a list of MemberId or Member objects."
        )
        raise ValueError(msg)

def _check_optional_member(m):
    """For `Client.update_autoproxy_settings`
    """
    if m is None: return None
    if isinstance(m, MemberId): return str(m)
    if isinstance(m, Member): return str(m.id)
    # otherwise
    try:
        if isinstance(m, Member):
            return str(m.id)
        elif isinstance(m, MemberId):
            return str(m)
        else:
            m = MemberId(m)
            return str(m)
    except (TypeError, ValueError):
        msg = (
            f"Could not cast {m!r} to a MemberId or Member. "
            f"Please pass in a MemberId or Member object."
        )
        raise ValueError(msg)


_PATCHABLE_SYSTEM_KEYS = {
    "name": _max_string_length("name", 100, null_allowed=False),
    "description": _max_string_length("description", 1000),
    "tag": _max_string_length("tag", 79),
    "pronouns": _max_string_length("pronouns", 100),
    # API will report whether any urls are publically (in)accessible
    "avatar_url": _max_string_length("avatar_url", 256),
    "banner": _max_string_length("banner", 256),
    "color": _check_color,
    "description_privacy": _check_privacy,
    "pronoun_privacy": _check_privacy,
    "member_list_privacy": _check_privacy,
    "group_list_privacy": _check_privacy,
    "front_privacy": _check_privacy,
    "front_history_privacy": _check_privacy,
}

_PATCHABLE_MEMBER_KEYS = {
    "name": _max_string_length("name", 100, null_allowed=False),
    "display_name": _max_string_length("display_name", 100),
    "color": _check_color,
    "birthday": _check_birthday,
    "pronouns": _max_string_length("pronouns", 100),
    "avatar_url": _max_string_length("avatar_url", 256),
    "banner": _max_string_length("banner", 256),
    "description": _max_string_length("description", 1000),
    "proxy_tags": _check_proxy_tags,
    "keep_proxy": bool,
    "visibility": _check_privacy,
    "name_privacy": _check_privacy,
    "description_privacy": _check_privacy,
    "birthday_privacy": _check_privacy,
    "pronoun_privacy": _check_privacy,
    "avatar_privacy": _check_privacy,
    "metadata_privacy": _check_privacy,
}

_PATCHABLE_GROUP_KEYS = {
    "name": _max_string_length("name", 100, null_allowed=False),
    "display_name": _max_string_length("display_name", 100),
    "description": _max_string_length("description", 1000),
    "icon": _max_string_length("icon", 256),
    "banner": _max_string_length("banner", 256),
    "color": _check_color,
    "name_privacy": _check_privacy,
    "description_privacy": _check_privacy,
    "icon_privacy": _check_privacy,
    "list_privacy": _check_privacy,
    "metadata_privacy": _check_privacy,
    "visibility": _check_privacy,
}

_PATCHABLE_SWITCH_KEYS = {
    "members": _check_members,
    "timestamp": _check_timestamp,
}

_PATCHABLE_AUTOPROXY_SETTINGS_KEYS = {
    "autoproxy_member": _check_optional_member,
}

_PATCHABLE_SYSTEM_SETTINGS_KEYS = {
    "timezone": _check_timezone,
    "pings_enabled": bool,
    "latch_timeout": lambda n: None if n is None else int(n), # time in seconds
    "member_default_private": bool,
    "group_default_private": bool,
    "show_private_info": bool,
}

_PATCHABLE_SYSTEM_GUILD_SETTINGS_KEYS = { # requires guild id
    "proxying_enabled": bool,
    "tag": _max_string_length("tag", 79),
    "tag_enabled": bool,
}

_PATCHABLE_MEMBER_GUILD_SETTINGS_KEYS = { # requires guild id
    "display_name": _max_string_length("display_name", 100),
    "avatar_url": _max_string_length("avatar_url", 256),
}

_PRIVACY_ASSOCIATED_KEYS = set((
    "visibility",
    "name_privacy",
    "description_privacy",
    "birthday_privacy",
    "pronoun_privacy",
    "avatar_privacy",
    "metadata_privacy",
    "member_list_privacy",
    "group_list_privacy",
    "front_privacy",
    "front_history_privacy",
    "icon_privacy",
    "list_privacy",
))
